/**
 * One feature template.
 * Also contain static methods for parsing entire template files.
 */

#include <config.h>

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <string>
#include "common.h"
#include "FeatureSegment.h"

FeatureSegment::FeatureSegment(Dictionary *dict, WordBuffer *wb, FeatureFunctions *func) {
	this->dict = dict;
	this->wb = wb;
	this->func = func;
}

/**
 * Construct FeatureSegment from raw data
 */

void FeatureSegment::init(int row, int col, int rowFunctionId, int colFunctionId, int transFunctionId) {
	this->row = row;
	this->col = col;
	this->rowFunctionId = rowFunctionId;
	this->colFunctionId = colFunctionId;
	this->transFunctionId = transFunctionId;
	this->value = DICTIONARY_NULL;
}

void FeatureSegment::init(int32_t *data) {
	row = (int)data[0];
	col = (int)data[1];
	rowFunctionId = (int)data[2];
	colFunctionId = (int)data[3];
	transFunctionId = (int)data[4];
	value = DICTIONARY_NULL;
}

// parse one feature segment
char *FeatureSegment::parse(char *token) {
	// parse one feature segment
	SKIPWS(token);
	if (*token++ != '%') {
		fprintf(stderr, "Expected '%%': %s\n", token-1);
		errno = 0;
		return NULL;
	}
	if (*token++ != '[') {
		fprintf(stderr, "Expected '[': %s\n", token-1);
		errno = 0;
		return NULL;
	}
	SKIPWS(token);
	char *old = token;
	int i = readInt(&token);
	if (old != token) {
		// row number
		row = i;
		rowFunctionId = FUNCTIONID_NULL;
	} else {
		// row function
		row = 0;
		char label[MAX_LINE_SIZE];
		int i = 0;
		while (i < MAX_LINE_SIZE-1 && *token && *token != ',') {
			if (!isspace(*token))
				label[i++] = *token;
			token++;
		}
		label[i] = '\0';
		rowFunctionId = func->convertFunctionId(label);
		if (rowFunctionId == FUNCTIONID_NULL) {
			fprintf(stderr, "Invalid row function name: %s\n", label);
			errno = 0;
			return NULL;
		}
		if (func->getFunction(rowFunctionId)->getType() != FUNCTION_ROW) {
			fprintf(stderr, "Invalid function: %s\n", label);
			errno = 0;
			return NULL;
		}
	}
	SKIPWS(token);
	if (*token++ != ',') {
		fprintf(stderr, "Expected ',': %s\n", token-1);
		errno = 0;
		return NULL;
	}
	SKIPWS(token);
	old = token;
	i = readInt(&token);
	if (old != token) {
		// column number
		col = i;
		colFunctionId = FUNCTIONID_NULL;
	} else {
		// function or column name
		char label[MAX_LINE_SIZE];
		int i = 0;
		while (i < MAX_LINE_SIZE-1 && *token && *token != ']') {
			if (!isspace(*token))
				label[i++] = *token;
			token++;
		}
		label[i] = '\0';

		if (i > 0 && label[i-1] != ')') {
			// column name
			col = wb->getHeaderIndex(label);
			if (col == ANY_COLUMN) {
				fprintf(stderr, "Invalid column name: %s\n", label);
				errno = 0;
				return NULL;
			}
			colFunctionId = FUNCTIONID_NULL;
		} else {
			// column function
			col = 0;
			colFunctionId = func->convertFunctionId(label);
			if (colFunctionId == FUNCTIONID_NULL) {
				fprintf(stderr, "Invalid column function name: %s\n", label);
				errno = 0;
				return NULL;
			}
			if (func->getFunction(colFunctionId)->getType() != FUNCTION_COLUMN) {
				fprintf(stderr, "Invalid function: %s\n", label);
				errno = 0;
				return NULL;
			}
		}
	}
	SKIPWS(token);
	if (*token++ != ']') {
		fprintf(stderr, "Expected ']': %s\n", token-1);
		errno = 0;
		return NULL;
	}
	SKIPWS(token);

	// transformation function
	if (*token && *token == '.') {
		token++;
		char label[MAX_LINE_SIZE];
		int i = 0;
		while (i < MAX_LINE_SIZE-1 && *token) {
			if (!isspace(*token))
				label[i++] = *token;
			token++;
		}
		label[i] = '\0';
		transFunctionId = func->convertFunctionId(label);
		if (func->getFunction(transFunctionId)->getType() != FUNCTION_TRANSFORM) {
			fprintf(stderr, "Invalid function: %s\n", label);
			errno = 0;
			return NULL;
		}
	} else {
		transFunctionId = FUNCTIONID_NULL;
	}

	return token;
}

string *FeatureSegment::toString(bool printValue) {
	string *result = new string();
	result->append("%[");
	if (rowFunctionId != FUNCTIONID_NULL) {
		string *s = func->getFunction(rowFunctionId)->toString();
		result->append(*s);
		delete s;
	} else {
		char s[MAX_LINE_SIZE];
		snprintf(s, sizeof(s), "%d", row);
		result->append(s);
	}
	if (colFunctionId != FUNCTIONID_NULL) {
		string *s = func->getFunction(colFunctionId)->toString();
		result->append(*s);
		delete s;
	} else {
		result->append(wb->getHeaderName(col));
	}
	result->append("]");
	if (printValue && value != DICTIONARY_UNKNOWN) {
		result->append("=");
		result->append(dict->getString(value));
	}
	return result;
}
