/**
 * One feature template.
 * Also contain static methods for parsing entire template files.
 */

#include <config.h>

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <string>
#include "common.h"
#include "FeatureTemplate.h"

/**
 * Construct FeatureTemplate from raw data
 */

FeatureTemplate::FeatureTemplate(Dictionary *dict, WordBuffer *wb, FeatureFunctions *func, vector<FeatureSegment*> *segments, tr1::unordered_map<FeatureSegment*, int, feature_segment_hash, feature_segment_equal> *segmentIndex) {
	this->dict = dict;
	this->wb = wb;
	this->func = func;
	this->segments = segments;
	this->segmentIndex = segmentIndex;

	segmentCount = 0;
	segmentIndices = NULL;
	featureData = NULL;
}

FeatureTemplate::~FeatureTemplate() {
	delete[] segmentIndices;
	delete[] featureData;
	delete[] tmpValues;

	// delete FeatureValue values of featureCount (they are allocated dynamicaly)
	// we can't do it directly
	vector<int*> delme;
	for (tr1::unordered_map<FeatureValue, int, feature_value_hash, feature_value_equal>::iterator iter = featureCount.begin(); iter != featureCount.end(); ++iter)
		delme.push_back(iter->first.values);
	featureCount.clear();
	for (vector<int*>::iterator iter = delme.begin(); iter != delme.end(); ++iter)
		delete[] *iter;
}

void FeatureTemplate::init(int count, int32_t *data) {
	segmentCount = count;
	segmentIndices = new int[segmentCount];
	tmpValues = new int[segmentCount];
	for (int i = 0; i < segmentCount; i++) {
		segmentIndices[i] = (int)data[i];
	}
	featureData = NULL;
}

// parse one line (create one feature template)
bool FeatureTemplate::parseFeatureTemplate(char *str) {
	FeatureSegment *fs = NULL;
	char *s = strdup(str);

	segmentCount = countChars(s, '/')+1;
	segmentIndices = new int[segmentCount];
	tmpValues = new int[segmentCount];

	char *helper = NULL;
	char *token = f_strtok_r(s, "/", &helper);
	int index = 0;
	while (token) {
		if (!fs)
			fs = new FeatureSegment(dict, wb, func);
		if (!fs->parse(token)) {
			delete fs;
			break;
		}
		tr1::unordered_map<FeatureSegment*, int, feature_segment_hash, feature_segment_equal>::iterator iter = segmentIndex->find(fs);
		if (iter == segmentIndex->end()) {
			int id = segmentIndex->size();
			segmentIndices[index] = id;
			(*segmentIndex)[fs] = id;
			fs = NULL;
		} else {
			segmentIndices[index] = iter->second;
			delete fs;
			fs = NULL;
		}
		token = f_strtok_r(NULL, "/", &helper);
		index++;
	}

	if (token) {
		delete[] segmentIndices;
		segmentIndices = NULL;
		delete[] tmpValues;
		tmpValues = NULL;
		segmentCount = 0;
		free(s);
		errno = 0;
		return false;
	}
	
	free(s);
	return true;
}

void FeatureTemplate::expandFeatureTemplate() {
	FeatureValue fv(segmentCount, tmpValues);
	getCurrentValue(&fv);
	tr1::unordered_map<FeatureValue, int, feature_value_hash, feature_value_equal>::iterator iter = featureCount.find(fv);
	if (iter == featureCount.end()) {
		int *values2 = new int[segmentCount];
		memcpy(values2, tmpValues, sizeof(int)*segmentCount);
		FeatureValue fv2(segmentCount, values2);
		featureCount[fv2] = 1;
	} else {
		iter->second++;
	}
}

// create features from feature counts
int FeatureTemplate::createFeatures(int minFreq, int totalFeatures) {
	int total = 0;
	for (tr1::unordered_map<FeatureValue, int, feature_value_hash, feature_value_equal>::iterator iter = featureCount.begin(); iter != featureCount.end(); ++iter) {
		if (iter->second >= minFreq)
			total++;
	}
	delete[] featureData;
	featureData = new int[total*segmentCount];
	featureIndex.rehash(total);
	int index = 0;
	for (tr1::unordered_map<FeatureValue, int, feature_value_hash, feature_value_equal>::iterator iter = featureCount.begin(); iter != featureCount.end(); ++iter) {
		if (iter->second >= minFreq) {
			int *values = iter->first.values;
			memcpy(featureData+index, values, sizeof(int)*segmentCount);
			FeatureValue fv(segmentCount, featureData+index);
			featureIndex[fv] = totalFeatures++;
			index += segmentCount;
		}
	}
	// delete FeatureValue values of featureCount (they are allocated dynamicaly)
	// we can't do it directly
	vector<int*> delme;
	for (tr1::unordered_map<FeatureValue, int, feature_value_hash, feature_value_equal>::iterator iter = featureCount.begin(); iter != featureCount.end(); ++iter)
		delme.push_back(iter->first.values);
	featureCount.clear();
	for (vector<int*>::iterator iter = delme.begin(); iter != delme.end(); ++iter)
		delete[] *iter;

	return totalFeatures;
}

int FeatureTemplate::addFeatures(int size, int32_t *data, int totalFeatures) {
	featureData = new int[size];
	for (int i = 0; i < size; i++) {
		featureData[i] = (int)data[i];
	}
	int n = size/segmentCount;
	featureIndex.rehash(n);
	for (int i = 0; i < n; i++) {
		FeatureValue fv(segmentCount, featureData+i*segmentCount);
		featureIndex[fv] = totalFeatures++;
	}
	return totalFeatures;
}

void FeatureTemplate::getFeatures(int *count, int **data) {
	*count = featureIndex.size();
	*data = featureData;
}

string *FeatureTemplate::toString(bool currentOnly) {
	string *result = new string();

	// dump template
	for (int i = 0; i < segmentCount; i++) {
		if (i > 0)
			result->append("/");
		else if (!currentOnly)
			result->append("# ");
		string *s = (*segments)[segmentIndices[i]]->toString(currentOnly);
		result->append(*s);
		delete s;
	}
	result->append("\n");

	if (!featureData || currentOnly)
		return result;

	// dump features
	for (int j = 0; j < (int)featureIndex.size(); j++) {
		for (int i = 0; i < segmentCount; i++) {
			if (i > 0)
				result->append("/");
			string *s = (*segments)[segmentIndices[i]]->toString(false);
			result->append(*s);
			delete s;
			result->append("=");
			result->append(dict->getString(featureData[j*segmentCount+i]));
		}
		result->append("\n");
	}
	return result;
}
