/**
 * Features
 */

#include <config.h>

#include <errno.h>
#include <string.h>
#include "Features.h"
#include "Dictionary.h"

bool Features::_loadFeatureTemplates(FILE *f) {
	char line[MAX_LINE_SIZE];
	bool eof = false;
	int index = 0;
	while (1) {
		do {
			errno = 0;
			if (!fgets(line, sizeof(line), f)) {
				if (errno > 0)
					return false;
				eof = true;
			}
		} while (!eof && (isEmpty(line) || isComment(line)));
		if (eof)
			break;
		chomp(line);

		// crate feature template
		FeatureTemplate *ft = new FeatureTemplate(dict, wb, func, &segments, &segmentIndex);
		if (!ft->parseFeatureTemplate(line)) {
			delete ft;
			errno = 0;
			return false;
		}
		featureTemplates.push_back(ft);

		index++;
	}

	int n = segmentIndex.size();
	segments.clear();
	segments.resize(n);
	for (tr1::unordered_map<FeatureSegment*, int, feature_segment_hash, feature_segment_equal>::iterator iter = segmentIndex.begin(); iter != segmentIndex.end(); ++iter) {
		segments[iter->second] = iter->first;
	}
	segmentIndex.clear();

	return true;
}

bool Features::_loadFeatures(FILE *f) {
	bool swap = false;
	// signature
	int32_t signature; 
        if (fread(&signature, sizeof(int32_t), 1, f) != 1)
		return false;
        if (signature != 0xbeef33) {
		signature = endian_swap(signature);
		if (signature != 0xbeef33) {
			errno = 0;
			return false;
		}
		swap = true;
	}

	// size of feature names
	int32_t size;
        if (fread(&size, sizeof(int32_t), 1, f) != 1)
		return false;
	if (swap)
		size = endian_swap(size);

	// load and compare feature names
	char *charData = new char[size];
        if (fread(charData, sizeof(char), size, f) != (unsigned)size) {
		delete[] charData;
		return false;
	}
	int index = 0;
	int id = 0;
	while (index < size) {
		if (strcmp(charData+index, wb->getHeaderName(id))) {
			fprintf(stderr, "Invalid column: %s (%s expected)\n", wb->getHeaderName(id), charData+index);
			delete [] charData;
			errno = 0;
			return false;
		}
		index += strlen(charData+index)+1;
		id++;
	}
	delete[] charData;

	// size of segments
        if (fread(&size, sizeof(int32_t), 1, f) != 1)
		return false;
	if (swap)
		size = endian_swap(size);
	int totalSegments = (int)size/5;
	assert(totalSegments < 10000000);
	for (vector<FeatureSegment*>::iterator iter = segments.begin(); iter != segments.end(); ++iter)
		delete *iter;
	segments.clear();

	// segments
	int32_t *data = new int32_t[size];
        if (fread(data, sizeof(int32_t), size, f) != (unsigned)size) {
		delete[] data;
		return false;
	}
	if (swap) {
		for (int i = 0; i < size; i++)
			data[i] = endian_swap(data[i]);
	}
	segments.resize(totalSegments);
	for (int i = 0; i < totalSegments; i++) {
		segments[i] = new FeatureSegment(dict, wb, func);
		segments[i]->init(data+5*i);
	}
	segmentIndex.clear();
	delete[] data;

	// size of feature functions
        if (fread(&size, sizeof(int32_t), 1, f) != 1)
		return false;
	if (swap)
		size = endian_swap(size);
	
	// feature function description
	charData = new char[size];
        if (fread(charData, sizeof(char), size, f) != (unsigned)size) {
		delete[] charData;
		return false;
	}
	index = 0;
	while (index < size) {
		func->convertFunctionId(charData+index);
		index += strlen(charData+index)+1;
	}
	delete[] charData;

	// size of feature templates
        if (fread(&size, sizeof(int32_t), 1, f) != 1)
		return false;
	if (swap)
		size = endian_swap(size);
	assert(size < 10000000);

	// feature templates
	data = new int32_t[size];
        if (fread(data, sizeof(int32_t), size, f) != (unsigned)size) {
		delete[] data;
		return false;
	}
	if (swap) {
		for (int i = 0; i < size; i++)
			data[i] = endian_swap(data[i]);
	}
	int i = 0;
	while (i < size) {
		int count = (int)data[i];
		FeatureTemplate *ft = new FeatureTemplate(dict, wb, func, &segments, &segmentIndex);
		ft->init(count, data+i+1);
		featureTemplates.push_back(ft);
		i += count+1;
	}
	delete[] data;

	// features
	for (int i = 0; i < (int)featureTemplates.size(); i++) {
		// size of features
        	if (fread(&size, sizeof(int32_t), 1, f) != 1)
			return false;
		if (swap)
			size = endian_swap(size);
		assert(size < 10000000);

		// feature values
		data = new int32_t[size];
        	if (fread(data, sizeof(int32_t), size, f) != (unsigned)size) {
			delete[] data;
			return false;
		}
		if (swap) {
			for (int i = 0; i < size; i++)
				data[i] = endian_swap(data[i]);
		}
		totalFeatures = featureTemplates[i]->addFeatures((int)size, data, totalFeatures);
		delete[] data;
	}
	return true;
}

bool Features::_saveFeatures(FILE *f, int minFreq) {
	// signature
	int32_t signature = 0xbeef33;
        if (fwrite(&signature, sizeof(int32_t), 1, f) != 1)
		return false;

	// size of feature column description
	int32_t size = 0;
	int columns = wb->getHeaderCount();
	for (int i = 0; i < columns; i++) {
		size += strlen(wb->getHeaderName(i))+1;
	}
        if (fwrite(&size, sizeof(int32_t), 1, f) != 1)
		return false;

	// feature columns description
	for (int i = 0; i < columns; i++) {
		const char *name = wb->getHeaderName(i);
		int len = strlen(name)+1;
        	if (fwrite(name, sizeof(char), len, f) != (unsigned)len)
			return false;
	}

	// size of segments
	size = segments.size()*5;
        if (fwrite(&size, sizeof(int32_t), 1, f) != 1)
		return false;

	// segments
	int32_t *data = new int32_t[size];
	for (int i = 0; i < (int)segments.size(); i++) {
		data[i*5+0] = (int32_t)segments[i]->row;
		data[i*5+1] = (int32_t)segments[i]->col;
		data[i*5+2] = (int32_t)segments[i]->rowFunctionId;
		data[i*5+3] = (int32_t)segments[i]->colFunctionId;
		data[i*5+4] = (int32_t)segments[i]->transFunctionId;
	}
        if (fwrite(data, sizeof(int32_t), size, f) != (unsigned)size) {
		delete[] data;
		return false;
	}
	delete[] data;

	// size of feature functions
	int nFunctions = func->getFunctionCount();
	const char **functionNames = new const char*[nFunctions];
	size = 0;
	for (int i = 0; i < nFunctions; i++) {
		functionNames[i] = func->getFunctionName(i);
		size += strlen(functionNames[i])+1;
	}
        if (fwrite(&size, sizeof(int32_t), 1, f) != 1) {
		delete[] functionNames;
		return false;
	}

	// feature function description
	for (int i = 0; i < nFunctions; i++) {
		unsigned int len = strlen(functionNames[i])+1;
        	if (fwrite(functionNames[i], sizeof(char), len, f) != len) {
			delete[] functionNames;
			return false;
		}
	}
	delete[] functionNames;

	// size of feature templates
	size = 0;
	for (int i = 0; i < (int)featureTemplates.size(); i++)
		size += featureTemplates[i]->getSegmentCount()+1;
        if (fwrite(&size, sizeof(int32_t), 1, f) != 1)
		return false;
	
	// feature templates
	data = new int32_t[size];
	int index = 0;
	for (int i = 0; i < (int)featureTemplates.size(); i++) {
		int count = featureTemplates[i]->getSegmentCount();
		data[index++] = (int32_t)count;
		int *indices = featureTemplates[i]->getSegmentIndices();
		for (int j = 0; j < count; j++) {
			data[index++] = (int32_t)indices[j];
		}
	}
        if (fwrite(data, sizeof(int32_t), size, f) != (unsigned)size) {
		delete[] data;
		return false;
	}
	delete[] data;

	// features
	for (int i = 0; i < (int)featureTemplates.size(); i++) {
		int featureTemplateCount;
		int *featureValues;
		featureTemplates[i]->getFeatures(&featureTemplateCount, &featureValues);
		int sc = featureTemplates[i]->getSegmentCount();
		size = (int32_t)featureTemplateCount*sc;
		
		// size of features
        	if (fwrite(&size, sizeof(int32_t), 1, f) != 1)
			return false;

		// features
		data = new int32_t[size];
		for (int j = 0; j < size; j++) {
			data[j] = (int32_t)featureValues[j];
		}
        	if (fwrite(data, sizeof(int32_t), size, f) != (unsigned)size) {
			delete[] data;
			return false;
		}
		delete[] data;
	}
	return true;
}

Features::Features(Dictionary *dict, WordBuffer *wb) {
	this->dict = dict;
	this->wb = wb;
	this->func = new FeatureFunctions(dict, wb);
	totalFeatures = 0;
	alphas = NULL;
	gammas = NULL;
	lastGammaUpdate = NULL;
	sentenceCount = 0;

	prevContextSize = NULL;
	nextContextSize = NULL;

	decAlphaIndex = NULL;
	incAlphaIndex = NULL;
}

Features::~Features() {
	delete func;

	for (vector<FeatureTemplate*>::iterator iter = featureTemplates.begin(); iter != featureTemplates.end(); ++iter) {
		delete *iter;
	}
	for (vector<FeatureSegment*>::iterator iter = segments.begin(); iter != segments.end(); ++iter) {
		delete *iter;
	}

	delete[] alphas;
	delete[] gammas;
	delete[] lastGammaUpdate;

	delete[] prevContextSize;
	delete[] nextContextSize;

	delete[] decAlphaIndex;
	delete[] incAlphaIndex;
}

void Features::init() {
	func->init();
	for (vector<FeatureTemplate*>::iterator iter = featureTemplates.begin(); iter != featureTemplates.end(); ++iter) {
		delete *iter;
	}
	featureTemplates.clear();
	totalFeatures = 0;
	for (vector<FeatureSegment*>::iterator iter = segments.begin(); iter != segments.end(); ++iter) {
		delete *iter;
	}
	segments.clear();
	segmentIndex.clear();

	delete[] alphas;
	alphas = NULL;
	delete[] gammas;
	gammas = NULL;
	delete[] lastGammaUpdate;
	lastGammaUpdate = NULL;
	sentenceCount = 0;

	delete[] prevContextSize;
	prevContextSize = NULL;
	delete[] nextContextSize;
	nextContextSize = NULL;

	delete[] decAlphaIndex;
	decAlphaIndex = NULL;
	delete[] incAlphaIndex;
	incAlphaIndex = NULL;
}

bool Features::loadFeatureTemplates(const char *file) {
	FILE *f = fopen(file, "r");
	if (!f)
		return false;

	if (!_loadFeatureTemplates(f)) {
		int err = errno;
		fclose(f);
		errno = err;
		return false;
	}
	featureTemplateCount = featureTemplates.size();

	fclose(f);

	return true;
}

bool Features::loadFeatures(const char *file) {
	FILE *f = fopen(file, "r");
	if (!f)
		return false;

	init();
	if (!_loadFeatures(f)) {
		int err = errno;
		fclose(f);
		errno = err;
		return false;
	}
	featureTemplateCount = featureTemplates.size();

	fclose(f);

	return true;
}

bool Features::saveExpandedFeatures(const char *file, int minFreq) {
	FILE *f = fopen(file, "w");
	if (!f)
		return false;

	for (vector<FeatureTemplate*>::iterator iter = featureTemplates.begin(); iter != featureTemplates.end(); ++iter) {
		totalFeatures = (*iter)->createFeatures(minFreq, totalFeatures);
	}
	if (!_saveFeatures(f, minFreq)) {
		int err = errno;
		fclose(f);
		errno = err;
		return false;
	}

	fclose(f);

	return true;
}

void Features::initAlphas() {
	delete[] decAlphaIndex;
	decAlphaIndex = new int[featureTemplateCount];
	delete[] incAlphaIndex;
	incAlphaIndex = new int[featureTemplateCount];

	delete[] prevContextSize;
	prevContextSize = new int[featureTemplateCount];
	delete[] nextContextSize;
	nextContextSize = new int[featureTemplateCount];

	sentenceCount = 0;
	alphas = new alpha_t[totalFeatures];
	gammas = new alpha_t[totalFeatures];
	lastGammaUpdate = new int[totalFeatures];
	for (int i = 0; i < totalFeatures; i++) {
		alphas[i] = 0;
		gammas[i] = 0;
		lastGammaUpdate[i] = -1;
	}
	sentenceId = 0;

}

void Features::finishAlphas() {
	for (int i = 0; i < totalFeatures; i++) {
		gammas[i] += alphas[i] * (sentenceId-1 - lastGammaUpdate[i]);
		lastGammaUpdate[i] = sentenceId;
	}
}

bool Features::loadAlphas(const char *file) {
	delete[] decAlphaIndex;
	decAlphaIndex = new int[featureTemplateCount];
	delete[] incAlphaIndex;
	incAlphaIndex = new int[featureTemplateCount];

	delete[] prevContextSize;
	prevContextSize = new int[featureTemplateCount];
	delete[] nextContextSize;
	nextContextSize = new int[featureTemplateCount];

	FILE *f = fopen(file, "r");
	if (!f)
		return false;

	// read signature (detect endianness)
	int32_t signature;
	bool swap = false;
	if (fread(&signature, sizeof(int32_t), 1, f) != 1)
		return false;
	if (signature != 0xbeef12+4 && signature != 0xbeef12+8) {
		signature = endian_swap(signature);
		swap = true;
	}
	switch (signature) {
	case 0xbeef12+4:
		if (ALPHA_SIZE != 32) {
			fprintf(stderr, "Incompatible file format (32bit)\n");
			errno = 0;
			return false;
		}
		break;
	case 0xbeef12+8:
		if (ALPHA_SIZE != 64) {
			fprintf(stderr, "Incompatible file format (64bit)\n");
			errno = 0;
			return false;
		}
		break;
	default:
		fprintf(stderr, "Invalid file format (unrecognized)\n");
		errno = 0;
		return false;
	}

	// read number of features
	uint32_t count;
	if (fread(&count, sizeof(uint32_t), 1, f) != 1)
		return false;
	if (swap)
		count = endian_swap(count);
	if (count != (unsigned)totalFeatures) {
		fprintf(stderr, "Number of features (%d) and alphas (%d) differ\n", totalFeatures, count);
		errno = 0;
		return false;
	}

	// read sentences count
	uint32_t count2;
	if (fread(&count2, sizeof(uint32_t), 1, f) != 1)
		return false;
	if (swap)
		count2 = endian_swap(count2);
	sentenceCount = count2;

	// read actual alphas
	alphas = new alpha_t[count];
	if (fread(alphas, sizeof(alpha_t), count, f) != count)
		return false;
	if (swap) {
		for (unsigned i = 0; i < count; i++)
			alphas[i] = endian_swap(alphas[i]);
	}

	// read actual gammas
	gammas = new alpha_t[count];
	if (fread(gammas, sizeof(alpha_t), count, f) != count)
		return false;
	if (swap) {
		for (unsigned i = 0; i < count; i++)
			gammas[i] = endian_swap(gammas[i]);
	}
	lastGammaUpdate = new int[count];
	for (int i = 0; i < (int)count; i++)
		lastGammaUpdate[i] = -1;
	sentenceId = 0;

	fclose(f);

	return true;
}

bool Features::saveAlphas(const char *file) {
	FILE *f = fopen(file, "w");
	if (!f)
		return false;

	// save signature
	int32_t signature = 0xbeef12+(ALPHA_SIZE/8);
	if (fwrite(&signature, sizeof(int32_t), 1, f) != 1)
		return false;

	// save features count
	uint32_t count = totalFeatures;
	if (fwrite(&count, sizeof(uint32_t), 1, f) != 1)
		return false;

	// save sentence count
	uint32_t count2 = sentenceCount;
	if (fwrite(&count2, sizeof(uint32_t), 1, f) != 1)
		return false;

	// save alphas
	if (fwrite(alphas, sizeof(alpha_t), count, f) != count)
		return false;

	// save gammas
	if (fwrite(gammas, sizeof(alpha_t), count, f) != count)
		return false;

	fclose(f);

//for (int i = 0; i < (int)features.size(); i++) {
//	PRINT_STRING(features[i]->toString());
//	fprintf(stderr, "\t%ld\t%ld\n", alphas[i], gammas[i]);
//}

	return true;
}

void Features::updateAlphas() {
	// update sentence count
	sentenceCount++;

	int sentenceSize = wb->getSentenceSize();
#ifdef DEBUG
	PRINT_STRING(wb->toString());
#endif
	for (int i = sentenceSize-1; i >= 0; i--) {
#ifdef DEBUG
		fprintf(stderr, "Word index: %d\n", i);
#endif

		// record proposed features for every template
		updateSegments(i, false);
		int j = 0;
		for (vector<FeatureTemplate*>::iterator iter = featureTemplates.begin(); iter != featureTemplates.end(); ++iter) {
			decAlphaIndex[j] = (*iter)->getCurrentFeatureIndex();
			j++;
		}

		// record correct features for every template
		updateSegments(i, true);
		j = 0;
		for (vector<FeatureTemplate*>::iterator iter = featureTemplates.begin(); iter != featureTemplates.end(); ++iter) {
			incAlphaIndex[j] = (*iter)->getCurrentFeatureIndex();
			j++;
		}
		// TODO: vyzkouset, jestli neni pole incAlphaIndex/decAlphaIndex uplne zbytecne: ze rychlost je stejna jako kdyby se delalo inc/dec pokazde
		for (int j = 0; j < featureTemplateCount; j++) {
			int dec = decAlphaIndex[j];
			int inc = incAlphaIndex[j];
			if (dec != inc) {
				if (dec >= 0) {
					if (lastGammaUpdate[dec] != sentenceId) {
						gammas[dec] += alphas[dec] * (sentenceId - lastGammaUpdate[dec]);
						lastGammaUpdate[dec] = sentenceId;
					}
					gammas[dec]--;
					alphas[dec]--;
				}

				if (inc >= 0) {
					if (lastGammaUpdate[inc] != sentenceId) {
						gammas[inc] += alphas[inc] * (sentenceId - lastGammaUpdate[inc]);
						lastGammaUpdate[inc] = sentenceId;
					}
					gammas[inc]++;
					alphas[inc]++;
				}
			}
		}
	}

	sentenceId++;
}

void Features::expandFeatureTemplates(int sentenceIndex) {
	updateSegments(sentenceIndex, true);
	for (vector<FeatureTemplate*>::iterator iter = featureTemplates.begin(); iter != featureTemplates.end(); ++iter) {
		(*iter)->expandFeatureTemplate();
	}
}

double Features::countScore(int sentenceIndex, bool useGamma) {
	double result = 0;

	// fill segments with current data
	updateSegments(sentenceIndex, false);

	for (vector<FeatureTemplate*>::iterator iter = featureTemplates.begin(); iter != featureTemplates.end(); ++iter) {

		// get index of current feature template values
		int index = (*iter)->getCurrentFeatureIndex();
		if (index >= 0) {
			if (useGamma && sentenceCount > 0)
				result = result + ((double)gammas[index]/sentenceCount);
			else
				result = result + alphas[index];
		}
	}

	return result;
}

double Features::countScore(int sentenceIndex, int prevContextSize, int nextContextSize, bool useGamma) {
/*	static int *prevContextSize = NULL;
	static int *nextContextSize = NULL;

	if (!prevSize) {
		int count = featureTemplates->size();
		prevSize = new int[count];
		nextSize = new int[count];
		for (int i = 0; i < count; i++) {
			
		}
	}*/
	// TODO

	double result = 0;

	// fill segments with current data
	updateSegments(sentenceIndex, false);

	for (vector<FeatureTemplate*>::iterator iter = featureTemplates.begin(); iter != featureTemplates.end(); ++iter) {

		// get index of current feature template values
		int index = (*iter)->getCurrentFeatureIndex();
		if (index >= 0) {
			if (useGamma && sentenceCount > 0)
				result = result + ((double)gammas[index]/sentenceCount);
			else
				result = result + alphas[index];
		}
	}

	return result;
}

void Features::getTagContextSize(int *min, int *max) {
	*min = 0;
	*max = 0;
	for (int i = 0; i < (int)segments.size(); i++) {
		segments[i]->getTagContextSize(min, max);
	}
}

void Features::updateSegments(int sentenceIndex, bool useCorrectTags) {
	int n = (int)segments.size();
	for (int i = 0; i < n; i++) {
		 segments[i]->update(sentenceIndex, useCorrectTags);
	}
}

string *Features::toString() {
	string *result = new string();

	for (vector<FeatureTemplate*>::iterator iter = featureTemplates.begin(); iter != featureTemplates.end(); ++iter) {
		string *s = (*iter)->toString();
		result->append(*s);
		result->append("\n");
		delete s;
	}

	return result;
}


