/**
 * Extended heap. This is a modified heap, so that we can remove any element
 * from it. Requires an objects that keeps track of the index itself (through
 * setHeapIndex()). Following methods are mandatory for objects within the
 * heap:
 *
 * number getScore()
 *
 * void setHeapIndex(int)
 *
 * string *toString()
 */

#ifndef _HEAP_H_
#define _HEAP_H_

#include <config.h>

#include <assert.h>
#include <string>
#include <vector>

template<class T>
class Heap {
	vector<T*> heap;
	int heapSize;

	int shiftdown(int index);
	void shiftup(int index);
public:
	Heap();
	~Heap() {};

	void clear();
	T *get(int index);
	T *getFirstUnused();
	void insertFirstUnused();
	void remove(int index);
	void insertFirstUnusedRaw();
	void rebuild();

	string *toString();
};

template<class T>
int Heap<T>::shiftdown(int index) {
	T *x = heap[index];
	int i, j;
	for (i = index; i > 1; i = j) {
		j = i >> 1;
		if (heap[j]->getScore() >= x->getScore())
			break;
		heap[i] = heap[j];
		heap[i]->setHeapIndex(i);
	}
	heap[i] = x;
	heap[i]->setHeapIndex(i);
	return i;
}

template<class T>
void Heap<T>::shiftup(int index) {
	int i = index;
	int j = index*2;
	while (j < heapSize) {
		if (j < heapSize-1 && heap[j]->getScore() < heap[j+1]->getScore())
			j++;
		if (heap[i]->getScore() >= heap[j]->getScore())
			break;
		T *x = heap[i];
		heap[i] = heap[j];
		heap[j] = x;
		heap[i]->setHeapIndex(i);
		heap[j]->setHeapIndex(j);
		i = j;
		j = j*2;
	}
}

template<class T>
Heap<T>::Heap() {
	heapSize = 0;
}

/**
 * Remove all elements from the heap (elements are not destroyed).
 */
template<class T>
void Heap<T>::clear() {
	heapSize = 0;
}

/**
 * Return element at given index.
 * @param index index of the heap to return.
 * @return Element with given index.
 */
template<class T>
T *Heap<T>::get(int index) {
	assert(index < heapSize);
	return heap[index];
}

/**
 * Return first element that is unused by the heap. If necessary, insert one
 * element at the end of the vector. This is used as a preparatory step before
 * insertFirstUnused().
 * @return First unused element.
 */
template<class T>
T *Heap<T>::getFirstUnused() {
	if ((unsigned)heapSize == heap.size())
		heap.push_back(new T());
	return heap[heapSize];
}

/**
 * Append first unused element into the heap. The element should be prepared before using getFirstUnused().
 */
template<class T>
void Heap<T>::insertFirstUnused() {
	heap[heapSize]->setHeapIndex(heapSize);
	heapSize++;
	shiftdown(heapSize-1);
}

/**
 * Remove an element from the heap.
 * @param index index of the element.
 */
template<class T>
void Heap<T>::remove(int index)
{
	if (index >= heapSize || heapSize < 0)
		return;
	if (index == heapSize-1) {
		heapSize--;
		return;
	}
	T *x = heap[index];
	heap[index] = heap[heapSize-1];
	heap[index]->setHeapIndex(index);
	heap[heapSize-1] = x;
	heapSize--;
	shiftup(index);
}

/**
 * Append new element at the end of the vector, but do not create a heap out of
 * the elements. This is a fast method, rebuild() should be called after
 * inserts are done.
 */
template<class T>
void Heap<T>::insertFirstUnusedRaw() {
	heapSize++;
}

/**
 * Compare two elements of a Heap using getScore() method.
 */
template<class T>
class CompareT {
public:
	/** Compare two elements of a Heap using getScore() method. */
	bool operator()(const T *x, const T *y) const {
		return x->getScore() < y->getScore();
	}
};

/**
 * Rebuild the heap. Used after (possibly many) insertFirstUnusedRaw() calls.
 */
template<class T>
void Heap<T>::rebuild() {
	make_heap(heap.begin(), heap.begin()+heapSize, CompareT<T>());
	for (int i = 0; i < heapSize; i++) {
		heap[i]->setHeapIndex(i);
	}
}

/**
 * Return a string representation of the heap.
 * The string should be deleted.
 */
template<class T>
string *Heap<T>::toString() {
	string *result = new string();
	unsigned br = 2;

	for (unsigned i = 1; i < heapSize; i++) {
		string *s = heap[i]->toString();
		result->append(*s);
		delete s;
		if (i+1 == br) {
			result->append("\n");
			br = br*2;
		} else {
			result->append(" ");
		}
	}
	result->append("\n");
	return result;
}

#endif
