/**
 * Span:
 */
#ifndef _SPAN_H_
#define _SPAN_H_

#include <config.h>

#include <limits>
#include <list>
#include <string>
#include "Features.h"
#include "Heap.h"
#include "WordBuffer.h"

using namespace std;

class Span {

	// one state of the span, contains only tagged words
	class State {
	public:
		// score of the top hypothesis (only for the head of a span)
		double topHypothesisScore;
		// tag for the top hypothesis
		int tagIndex;

		State() {};
		~State() {};
		void set(double topHypothesisScore, int tagIndex);
		double getTopHypothesisScore();
		int getTagIndex();
		string *toString();
	};

	// non-tagged words
	class Hypothesis {
		double hypothesisScore;
		double actionScore;
		// tag at the current position
		int tagIndex;
		// index of left state of this hypothesis (-1 = no left state)
		int leftStateIndex;
		// index of right state of this hypothesis (-1 = no right state)
		int rightStateIndex;
	public:
		Hypothesis() {};
		~Hypothesis() {};
		void set(double hypothesisScore, double actionScore, int tagIndex, int leftStateIndex, int rightStateIndex);
		double getHypothesisScore() const;
		double getActionScore() const;
		int getLeftStateIndex() const;
		int getRightStateIndex() const;
		// for heap ordering
		bool operator<(const Hypothesis &h) const;
	};

	// used by a hash_set for detection of already selected states
	class HypothesisInterface {
	public:
		int *tags;
		int count;

		HypothesisInterface() {};
		~HypothesisInterface() {};
		void set(Span *span, int hypothesisIndex);
	};
	struct hypothesis_interface_hash: public unary_function<HypothesisInterface*, size_t> {
		size_t operator() (const HypothesisInterface *hi) const {
			size_t result = 0;
			for (int i = 0; i < hi->count; i++)
				result = result*13 + hi->tags[i];
			return tr1::hash<int>()(result);
		}
	};
	struct hypothesis_interface_equal {
		bool operator()(const HypothesisInterface *hi1, const HypothesisInterface *hi2) const {
			assert(hi1->count == hi2->count);
			return !memcmp(hi1->tags, hi2->tags, hi1->count);
		}
	};

	// action score heap item
	class ActionScore {
		int sentenceIndex;
		double score;
		int heapIndex;
	public:
		ActionScore() {};
		~ActionScore() {};
		void set(int sentenceIndex, double score);
		int getSentenceIndex() const;
		int getHeapIndex() const;
		double getScore() const;	// required by Heap
		void setHeapIndex(int index);	// required by Heap
	};

	// used features: to improve performance of training
	class UsedFeature {
		list<UsedFeature>::iterator iter;
		int sentenceIndex;
		int hypothesisIndex;
	public:
		UsedFeature() {};
		~UsedFeature() {};
	};

	// this word was already tagged
	bool tagged;
	// first word of a span
	int firstWordIndex;
	// last word of a span (-1 if this instance is not head of a span)
	int lastWordIndex;

	// tagged words: we keep n best states
	State *states;
	int stateCount;

	// untagged words: we keep n best hypotheses
	vector<Hypothesis*> hypotheses;
	int hypothesisCount;
	// pointer to best action score
	ActionScore *bestActionScore;

	// beam width, common to all spans
	static int beamWidth;

	// size of span interface
	static int interfaceSize;

	// spans for the whole sentence
	static vector<Span*> *spans;

	// record used features of every hypothesis to improve the speed of training
	static tr1::unordered_map<int, list<UsedFeature> > *usedFeatures;

	// heap of action scores in the whole sentence
	static Heap<ActionScore> actionScoreHeap;

	static WordBuffer *wb;
public:
	static void Init(WordBuffer *wb, vector<Span*> *spans, int beamWidth, int interfaceSize);
	static void Destroy();

	Span(int sentenceIndex);
	~Span();
	void reset(int sentenceIndex);

	bool isTagged() const;
	int getFirstWordIndex() const;
	int getLastWordIndex() const;
	int getSpanLength() const;
	int getStateCount() const;
	Span::State *getState(int index) const;
	double getTopActionScore() const;
	Span *getPreviousSpan() const;
	Span *getFollowingSpan() const;
	bool compatibleStateTag(int index1, int index2) const;
	bool equivalentStates(int leftStateIndex1, int leftStateIndex2, int rightStateIndex1, int rightStateIndex2);
	int fillTags(int sentenceIndex, int stateIndex, bool leftInterface);
	void updateScore(Features *ftrs);
	void tagSpan();

	string *toString();

	static void initSentence(int sentenceSize);
	static int getTopActionScoreIndex(bool buildHash);
};

// State

inline double Span::State::getTopHypothesisScore() {
	return topHypothesisScore;
}

inline int Span::State::getTagIndex() {
	return tagIndex;
}


// Hypothesis

inline void Span::Hypothesis::set(double hypothesisScore, double actionScore, int tagIndex, int leftStateIndex, int rightStateIndex) {
       this->hypothesisScore = hypothesisScore;
       this->actionScore = actionScore;
       this->tagIndex = tagIndex;
       this->leftStateIndex = leftStateIndex;
       this->rightStateIndex = rightStateIndex;
}

inline double Span::Hypothesis::getHypothesisScore() const {
       return hypothesisScore;
}

inline double Span::Hypothesis::getActionScore() const {
       return actionScore;
}

inline int Span::Hypothesis::getLeftStateIndex() const {
       return leftStateIndex;
}

inline int Span::Hypothesis::getRightStateIndex() const {
       return rightStateIndex;
}

inline bool Span::Hypothesis::operator<(const Hypothesis &h) const {
       return hypothesisScore < h.getHypothesisScore();
}

// ActionScore

inline void Span::ActionScore::set(int sentenceIndex, double score) {
       this->sentenceIndex = sentenceIndex;
       this->score = score;
}

inline int Span::ActionScore::getSentenceIndex() const {
       return sentenceIndex;
}

inline int Span::ActionScore::getHeapIndex() const {
       return heapIndex;
}

inline double Span::ActionScore::getScore() const {
       return score;
}

inline void Span::ActionScore::setHeapIndex(int index) {
       heapIndex = index;
}

// Span

inline bool Span::isTagged() const {
       return tagged;
}

inline int Span::getFirstWordIndex() const {
       return firstWordIndex;
}

inline int Span::getLastWordIndex() const {
       return lastWordIndex;
}

inline int Span::getSpanLength() const {
       return lastWordIndex-firstWordIndex+1;
}

inline int Span::getStateCount() const {
       return stateCount;
}

inline Span::State *Span::getState(int index) const {
       return &states[index];
}

inline double Span::getTopActionScore() const {
       return bestActionScore ? bestActionScore->getScore() : std::numeric_limits<double>::min();
}

inline Span *Span::getPreviousSpan() const {
       return firstWordIndex-1 >= 0 ? (*spans)[firstWordIndex-1] : NULL;
}


inline Span *Span::getFollowingSpan() const {
       return lastWordIndex+1 < (int)spans->size() ? (*spans)[lastWordIndex+1] : NULL;
}


inline bool Span::compatibleStateTag(int index1, int index2) const {
       return states[index1].tagIndex == states[index2].tagIndex;
}


#endif

// veta: vector spanu (roste tak, jak roste delka vety... stejne jako u Viterbiho)
//
// kazdy span:
//	- jestli uz jsem otagovany, ano:
//		- int firstWord, lastWord: odkud kam je aktualni span (a firstWord+?? pro odloucene spany?)
//		- n stavu (n x)
//			- top hypoteza (jakoze kompletni tagovani, tady jeden tag, zbytek v odloucenych spanech)
//			- skore top hypotezy (stejne cislo pro cely span???)
//	- ne:
//		- n nejlepsich top hypotez (pro kazdy stav jedna -- heap podle skore top hypotezy):
//			- index stavu nelevo a napravo top hypotezy a souvisejici action skore
