/**
 * Viterbi
 */

#include <config.h>

#include <string.h>
#include <assert.h>
#include <algorithm>
#include "Viterbi.h"

Viterbi::Viterbi(Dictionary *dict, WordBuffer *wb) {
	this->dict = dict;
	this->wb = wb;
	this->order = 3;
	this->prevTags = NULL;
	this->prune = 0;
	this->nbest = 0;
	pruneHeap = NULL;
}

Viterbi::~Viterbi() {
	for (vector<NBestState*>::iterator iter = statesNBest.begin(); iter != statesNBest.end(); ++iter) {
		delete *iter;
	}
	delete[] prevTags;
	delete[] pruneHeap;
}

void Viterbi::init(int prune, int nbest, int order) {
	delete[] prevTags;
	delete[] pruneHeap;
	states.clear();
	for (vector<NBestState*>::iterator iter = statesNBest.begin(); iter != statesNBest.end(); ++iter) {
		delete *iter;
	}
	statesNBest.clear();

	this->order = order;
	this->prevTags = NULL;
	this->prune = prune;
	this->nbest = (nbest == 0 || nbest == 1) ? 0 : nbest;
	pruneHeap = prune > 0 ? new DoubleInt[prune] : NULL;
}

// set previous tags for the given starting point [sentenceIndex, index]
void Viterbi::setPrevProposedTags(int index, int count, int offset) {
	for (int i = 0; i < count && index > 0; i++) {
		wb->setProposedTagIndex(offset-i, states[index].tagIndex);
		index = states[index].back;
	}
}

// save nbest for a state
void Viterbi::SaveStateNBest(int state, int back, double score) {
	NBestState *nbs = statesNBest[state];
	DoubleInt *s = nbs->states;

	// we did not reach n-best states yet
	if (nbs->count < nbest) {
		s[nbs->count].i = back;
		s[nbs->count].d = score;
		nbs->count++;
		if (nbs->count == nbest)
			make_heap(s, s+nbest);
		return;
	}

	// state is too cheap
	if (score < s[0].d)
		return;

	// overwrite lowest-scoring state
	pop_heap(s, s+nbest);
	s[nbest-1].i = back;
	s[nbest-1].d = score;
	push_heap(s, s+nbest);
}

/*
First possibility of storing states, used in processSentencePlain()

0	AB	ACD	AB	DA	ABC	A	B	A

3-grams:

00	0A	AA	AA	AD	DA	AA	AB	BA
	0B	AC	CA	AA	DB	BA
		AD	DA	BD	DC	CA
		BA	AB	BA	AA
		BC	CB		AB
		BD	DB		AC

4-grams:

000	00A	0AA	AAA	AAD	ADA	DAA	AAB	ABA
	00B	0AC	AAB	AAA	ADB	DBA	BAB
		0AD	ACA	ABD	ADC	DCA	CAB
		0BA	ACB	ABA	AAA	AAA
		0BC	ADA	CAD	AAB	ABA
		0BD	ADB	CAA	AAC	ACA
			BAA	CBD	BDA
			BAB	CBA	BDB
			BCA	DAD	BDC
			BCB	DAA	BAA
			BDA	DBD	BAB
			BDB	DBA	BAC
*/

/* no prunning, fastest method, compatible with Morce */
void Viterbi::processSentencePlain(Features *ftrs, int prevTagCount, bool useGamma) {
	// count size of ViterbiState array
	int sentenceSize = wb->getSentenceSize();
	int statesSize = 1;
	int currWordStates = 1;
	for (int i = 0; i < sentenceSize; i++) {
		int oldOfferSize = wb->getOfferSize(i-order+1);
		assert(oldOfferSize != 0);
		currWordStates /= oldOfferSize == -1 ? 1 : oldOfferSize;
		currWordStates *= wb->getOfferSize(i);
		statesSize += currWordStates;
	}
	states.resize(statesSize);
	for (int i = 1; i < statesSize; i++) {
		states[i].tagIndex = -1;
	}

	// init n-best states
	if (nbest > 0) {
		int size = statesNBest.size();
		for (int i = 0; i < statesSize; i++) {
			if (i >= size) {
				NBestState *s = new NBestState(nbest);
				statesNBest.push_back(s);
			} else {
				statesNBest[i]->count = 0;
			}
		}
	}

	// init Viterbi pass
	states[0].tagIndex = -1;
	states[0].back = 0;
	states[0].best = 0;

	// number of states for current sentence index
	// (offerSize(-2)*offerSize(-1)*offerSize(0) in case of 3-grams)
	currWordStates = 1;
	int prevFirst = 0;
	int prevLast = 0;
	// for every word in a sentence
	for (int sentenceIndex = 0; sentenceIndex < sentenceSize; sentenceIndex++) {
		int currFirst = prevLast + 1;
		int currLast = currFirst - 1;
		int currOfferSize = wb->getOfferSize(sentenceIndex);

		// lower number of states (too old history for current order of the model)
		int oldOfferSize = wb->getOfferSize(sentenceIndex-order+1);
		currWordStates /= oldOfferSize == -1 ? 1 : oldOfferSize;

		// for every previous state
		int prevTagsIndex = 0;
		for (int prevIndex = prevFirst; prevIndex <= prevLast; prevIndex++) {
			if (sentenceIndex > 0)
				setPrevProposedTags(prevIndex, sentenceIndex >= prevTagCount ? prevTagCount : sentenceIndex, sentenceIndex-1);

			// for every TAG at current possition
			for (int currTagIndex = 0; currTagIndex < currOfferSize; currTagIndex++) {
				int index = currFirst + prevTagsIndex * currOfferSize + currTagIndex;
				if (index > currLast)
					currLast = index;

				wb->setProposedTagIndex(sentenceIndex, currTagIndex);

				// count score	
				double score = ftrs->countScore(sentenceIndex, useGamma);
				if (states[index].tagIndex == -1 || states[prevIndex].best + score >= states[index].best) {
					states[index].tagIndex = currTagIndex;
					states[index].back = prevIndex;
					states[index].best = states[prevIndex].best + score;
				}
				// save n top-scoring states
				if (nbest > 0)
					SaveStateNBest(index, prevIndex, states[prevIndex].best + score);
			}

			prevTagsIndex = (prevTagsIndex + 1) % currWordStates;
		}
		// count current offer size too
		currWordStates *= currOfferSize;

		prevFirst = currFirst;
		prevLast = currLast;
	}
	lastFirst = prevFirst;
	lastLast = prevLast;
}

void Viterbi::Prune(int &currFirst, int &currLast) {
	if (currLast-currFirst+1 <= prune)
		return;

	// construct pruneHeap
	for (int i = 0; i < prune; i++) {
		pruneHeap[i].i = currFirst+i;
		pruneHeap[i].d = states[currFirst+i].best;
	}
	make_heap(pruneHeap, pruneHeap+prune);

	// go through all states, insert them into the heap
	for (int i = prune; i < currLast-currFirst+1; i++) {
		if (states[currFirst+i].best > pruneHeap[0].d) {
			pop_heap(pruneHeap, pruneHeap+prune);
			// mark cheapest state
			states[pruneHeap[prune-1].i].tagIndex = -1;
			pruneHeap[prune-1].i = currFirst+i;
			pruneHeap[prune-1].d = states[currFirst+i].best;
			push_heap(pruneHeap, pruneHeap+prune);
		} else {
			states[currFirst+i].tagIndex = -1;
		}
	}
	
	// eliminate cheap states
	int free = currFirst;
	for (int i = currFirst; i <= currLast; i++) {
		// copy valuable state over latest invaluable
		if (states[i].tagIndex != -1) {
			if (i != free) {
				memcpy(&states[free], &states[i], sizeof(ViterbiState));
				if (nbest > 1) {
					NBestState *s = statesNBest[free];
					statesNBest[free] = statesNBest[i];
					statesNBest[i] = s;
				}
			}
			free++;
		}
	}
	// zero counts for freed nbest states
	if (nbest > 1) {
		for (int i = free; i <= currLast; i++)
			statesNBest[i]->count = 0;
	}

	assert(free-currFirst == prune);
	currLast = free-1;
}

/*
Second possibility of storing states, used in processSentencePrune()

0	AB	ACD	AB	DA	ABC	A	B	A

3-grams:

00	0A	AA	AA	AD	DA	AA	AB	BA
	0B	BA	CA	BD	AA	BA
		AC	DA	AA	DB	CA
		BC	AB	BA	AB
		AD	CB		DC
		BD	DB		AC
4-grams:

000	00A	0AA	AAA	AAD	ADA	DAA	AAB	ABA
	00B	0BA	BAA	CAD	BDA	AAA	BAB
		0AC	ACA	DAD	AAA	DBA	CAB
		0BC	BCA	ABD	BAA	ABA
		0AD	ADA	CBD	ADB	DCA
		0BD	BDA	DBD	BDB	ACA
			AAB	AAA	AAB
			BAB	CAA	BAB
			ACB	DAA	ADC
			BCB	ABA	BDC
			ADB	CBA	AAC
			BDB	DBA	BAC
			BDB	DBA	BAC
*/

/* the same as plain, but prune states */
void Viterbi::processSentencePrune(Features *ftrs, int prevTagCount, bool useGamma) {
	// init prevTags if not already done
	if (prevTags == NULL && order > 2)
		prevTags = new int[order-2];

	// count size of ViterbiState array
	int sentenceSize = wb->getSentenceSize();
	int statesSize = 1;
	int currWordStates = 1;
	int maxStateSize = 0;
	for (int i = 0; i < sentenceSize; i++) {
		int oldOfferSize = wb->getOfferSize(i-order+1);
		currWordStates /= oldOfferSize == -1 ? 1 : oldOfferSize;
		currWordStates *= wb->getOfferSize(i);
		int size = currWordStates;

		if (size > maxStateSize)
			maxStateSize = size;
		if (prune < size)
			size = prune;
		statesSize += size;
	}
	if (prune < maxStateSize)
		statesSize += (maxStateSize - prune);

	states.resize(statesSize);

	// init n-best states
	if (nbest > 0) {
		int size = statesNBest.size();
		for (int i = 0; i < statesSize; i++) {
			if (i >= size) {
				NBestState *s = new NBestState(nbest);
				statesNBest.push_back(s);
			} else {
				statesNBest[i]->count = 0;
			}
		}
	}

	// init Viterbi pass
	states[0].tagIndex = -1;
	states[0].back = 0;
	states[0].best = 0;

	int prevFirst = 0;
	int prevLast = 0;
	// for every word in a sentence
	for (int sentenceIndex = 0; sentenceIndex < sentenceSize; sentenceIndex++) {
		// for every TAG at current possition
		int currFirst = prevLast + 1;
		int currLast = currFirst - 1;
		int offerSize = wb->getOfferSize(sentenceIndex);
		for (int currTagIndex = 0; currTagIndex < offerSize; currTagIndex++) {
			for (int t = 0; t < order-2; t++)
				prevTags[t] = -1;
			bool first = true;

			// for every previous state
			for (int i = prevFirst; i <= prevLast; i++) {
				if (sentenceIndex > 0)
					setPrevProposedTags(i, sentenceIndex >= prevTagCount ? prevTagCount : sentenceIndex, sentenceIndex-1);
				wb->setProposedTagIndex(sentenceIndex, currTagIndex);

				// count score	
				double score = ftrs->countScore(sentenceIndex, useGamma);

				// find whether previous tags are new (state should be changed)	
				// does not work for order == 2, so we test it separately later
				bool changed = false;
				int si = i;
				for (int t = 0; t < order-2 && si != 0; t++) {
					if (prevTags[t] != states[si].tagIndex) {
						changed = true;
						prevTags[t] = states[si].tagIndex;
					}
					si = states[si].back;
				}
				if (first) {
					changed = true;
					first = false;
				}
				if (changed) {
					// new state
					currLast++;
					states[currLast].tagIndex = currTagIndex;
					states[currLast].back = i;
					states[currLast].best = states[i].best + score;
				} else if (states[i].best + score >= states[currLast].best) {
					// update state
					states[currLast].back = i;
					states[currLast].best = states[i].best + score;
				}

				// save n top-scoring states
				if (nbest > 0)
					SaveStateNBest(currLast, i, states[i].best + score);
			}
		}
		Prune(currFirst, currLast);
		prevFirst = currFirst;
		prevLast = currLast;
	}
	lastFirst = prevFirst;
	lastLast = prevLast;
}

void Viterbi::processSentence(Features *ftrs, int prevTagCount, bool useGamma) {
	if (prune == 0) {
		processSentencePlain(ftrs, prevTagCount, useGamma);
	} else {
		processSentencePrune(ftrs, prevTagCount, useGamma);
	}
}

void Viterbi::bestSentenceTags() {
	int sentenceSize = wb->getSentenceSize();

	// find max in the last column
	int back = states[lastFirst].back;
	int tagIndex = states[lastFirst].tagIndex;
	double best = states[lastFirst].best;
	for (int i = lastFirst+1; i <= lastLast; i++) {
		if (states[i].best > best) {
			tagIndex = states[i].tagIndex;
			best = states[i].best;
			back = states[i].back;
		}
	}
	assert(tagIndex != -1);
	wb->setProposedTagIndex(sentenceSize-1, tagIndex);

	// process backward, fill proposed tag values
	for (int i = 1; i < sentenceSize; i++) {
		wb->setProposedTagIndex(sentenceSize-1-i, states[back].tagIndex);
		back = states[back].back;
	}
	assert(back == 0);
}

struct StackItem {
	int stateIndex;
	int NBestIndex;
	int sentenceIndex;
	double scoreFollowing;
};

// size of result must be sentenceSize x nbest
#define ZERO 1e-6
void Viterbi::bestSentenceTagsNBest(double *score) {
	assert(nbest > 1);
	int sentenceSize = wb->getSentenceSize();
	// result (sentenceSize x nbest)
	int *result = new int[nbest*sentenceSize];
	// score of this result and index in the result array
	DoubleInt *resultHeap = new DoubleInt[nbest];
	resultHeap[0].d = 0;
	// current size of resultHeap
	int resultCount = 0;
	// stack for back-tracking
	vector<StackItem> stack;
	// current tags of the sentence as we do backtracking
	int *currentTags = new int[sentenceSize];
	// heap of n best scoring hypotheses
	DoubleInt *bestHeap = new DoubleInt[nbest];
	// current size of bestHeap
	int bestCount = 0;

	// we do a trick here: in order to enable optimization (*) so that we
	// do not need to test all paths, we have to eliminate negative scores.
	// So we find the lowest negative number and subtract it from all
	// scores for given sentenceIndex.
	double totalCorrection = 0;
	double min = 0;
	int state = 1;
	int first = 1;
	for (int i = 0; i < sentenceSize; i++) {
		min = 0;
		int save = state;
		while (state < lastLast+1 && states[state].back < first) {
			double score = states[state].best - (states[states[state].back].best - totalCorrection);
			if (score < min)
				min = score;
			for (int j = 0; j < statesNBest[state]->count; j++) {
				DoubleInt *st = statesNBest[state]->getState(j);
				double score = st->d - (states[st->i].best - totalCorrection);
				if (score < min)
					min = score;
			}
			state++;
		}

		totalCorrection -= min;
		state = save;
		while (state < lastLast+1 && states[state].back < first) {
			states[state].best += totalCorrection;
			for (int j = 0; j < statesNBest[state]->count; j++) {
				statesNBest[state]->getState(j)->d += totalCorrection;
			}
			state++;
		}

		first = state;
	}

	// find n top-scoring final states
	assert(lastFirst <= lastLast);
	for (int i = lastFirst; i <= lastLast; i++) {
		NBestState *nbs = statesNBest[i];
		for (int j = 0; j < nbs->count; j++) {
			double score = nbs->getState(j)->d;
			if (bestCount < nbest) {
				// first n states
				bestHeap[bestCount].d = score;
				bestHeap[bestCount].i = i*nbest+j;
				bestCount++;
				if (bestCount == nbest)
					make_heap(bestHeap, bestHeap+nbest);
			} else {
				// state is expensive enough?
				if (score > bestHeap[0].d) {
					// overwrite lowest-scoring state
					pop_heap(bestHeap, bestHeap+nbest);
					bestHeap[nbest-1].d = score;
					bestHeap[nbest-1].i = i*nbest+j;
					push_heap(bestHeap, bestHeap+nbest);
				}
			}
		}
	}

	for (int i = bestCount-1; i >= 0; i--) {
		// save one of best state into the stack
		int lastState = bestHeap[i].i / nbest;
		int lastNBestIndex = bestHeap[i].i % nbest;
		DoubleInt *di = statesNBest[lastState]->getState(lastNBestIndex);
		int prevState = di->i;
		double lastScore = di->d;
		currentTags[sentenceSize-1] = states[lastState].tagIndex;
		if (sentenceSize == 1) {
			// we are finished with the sentence
			int index;
			if (resultCount == nbest) {
				// overwrite lowest-scoring result
				pop_heap(resultHeap, resultHeap+nbest);
				resultHeap[nbest-1].d = bestHeap[i].d;
				index = resultHeap[nbest-1].i;
				push_heap(resultHeap, resultHeap+nbest);
			} else {
				resultHeap[resultCount].i = resultCount;
				resultHeap[resultCount].d = bestHeap[i].d;
				index = resultCount;
				resultCount++;
				if (resultCount == nbest)
					make_heap(resultHeap, resultHeap+nbest);
			}
			memcpy(result+sentenceSize*index, currentTags, sizeof(int)*sentenceSize);
			continue;
		}
		
		// stateIndex, NBestIndex, sentenceIndex, scoreFollowing
		StackItem si = { prevState, statesNBest[prevState]->count-1, sentenceSize-2, lastScore-states[prevState].best };
		stack.push_back(si);
		// do a back-tracking for this state, try to find better paths
		while (stack.size() > 0) {
			// remove last item from the stack, put it into current state
			StackItem cs = stack.back();
			stack.pop_back();
			// for all n values 
			while (cs.NBestIndex >= 0) {
				NBestState *nbs = statesNBest[cs.stateIndex];
				DoubleInt *state = nbs->getState(cs.NBestIndex);
				double score = state->d + cs.scoreFollowing;
				// can we improve the result? (*)
				if ((bestCount == nbest && score+ZERO < bestHeap[0].d) || (resultCount == nbest && score-ZERO < resultHeap[0].d)) {
					// no, too low score
					cs.NBestIndex--;
					continue;
				}
				// yes, we can improve result, record tag
				currentTags[cs.sentenceIndex] = states[cs.stateIndex].tagIndex;
				// do we have a complete sentence?
				if (cs.sentenceIndex == 0) {
					// yes: copy tags to result, insert into resultHeap
					int index;
					if (resultCount == nbest) {
						// overwrite lowest-scoring result
						pop_heap(resultHeap, resultHeap+nbest);
						resultHeap[nbest-1].d = score;
						index = resultHeap[nbest-1].i;
						push_heap(resultHeap, resultHeap+nbest);
					} else {
						resultHeap[resultCount].i = resultCount;
						resultHeap[resultCount].d = score;
						index = resultCount;
						resultCount++;
						if (resultCount == nbest)
							make_heap(resultHeap, resultHeap+nbest);
					}
					memcpy(result+sentenceSize*index, currentTags, sizeof(int)*sentenceSize);
					cs.NBestIndex--;
				} else {
					// no: save state on the stack and create the new one
					if (--cs.NBestIndex >= 0)
						stack.push_back(cs);
					cs.scoreFollowing += state->d-states[state->i].best;
					cs.stateIndex = state->i;
					cs.NBestIndex = statesNBest[cs.stateIndex]->count-1;
					cs.sentenceIndex--;
				}
			}
		}
	}

	// sort result
	if (resultCount < nbest)
		make_heap(resultHeap, resultHeap+resultCount);
	sort_heap(resultHeap, resultHeap+resultCount);

	// save result to WordBuffer
	for (int i = 0; i < sentenceSize; i++) {
		for (int j = 0; j < nbest; j++) {
			int tagIndex = j >= resultCount ? -1 : result[resultHeap[j].i*sentenceSize+i];
			wb->setProposedTagIndex(i, tagIndex, j);
		}
	}

	// save scores
	if (score != NULL) {
		for (int i = 0; i < nbest; i++) {
			score[i] = i < resultCount ? resultHeap[i].d - totalCorrection : .0;
		}
	}
	delete[] bestHeap;
	delete[] currentTags;
	delete[] resultHeap;
	delete[] result;
}

string *Viterbi::toString() {
	string *result = new string();

	int sentenceSize = wb->getSentenceSize();

	result->append("Viterbi:\n");
	int state = 1;
	int first = 1;
	for (int i = 0; i < sentenceSize; i++) {
		while (state < lastLast+1 && states[state].back < first) {
			// print ViterbiState
			char s[1024];
			snprintf(s, sizeof(s), " [%d: %s %.3lf ->%d", state,  dict->getString(wb->getProposedTagValue(states[state].tagIndex)), states[state].best, states[state].back);
			result->append(s);
			if (nbest > 1) {
				result->append(" (");
				for (int i = 0; i < statesNBest[state]->count; i++) {
					if (i > 0)
						result->append(" ");
					DoubleInt *st = statesNBest[state]->getState(i);
					snprintf(s, sizeof(s), "%.3lf ->%d", st->d, st->i);
					result->append(s);
				}
				result->append(")");
			}
			result->append("]");

			state++;
		}
		first = state;
		result->append("\n");
	}

	return result;
}
