/** @file */
#ifndef _VITERBI_H_
#define _VITERBI_H_

#include <config.h>

#include <assert.h>
#include <string>
#include <vector>
#include "Features.h"
#include "WordBuffer.h"
#include "Dictionary.h"

using namespace std;

/**
 * Viterbi algorithm implementation. Prunning and n-best are implemented, as
 * well as arbitrary size of the model (e.g. you can have 2-gram, 3-gram,
 * 4-gram models).
 */
class Viterbi {

	/**
	 * This class is used for storing Double-ordered items in a heap
	 * the heap is sorted so that head is the smallest number
	 */
	class DoubleInt {
	public:
		double d;
		int i;

		// this is intentionaly incorrect to enable heap to be sorted in other
		// direction
		bool operator<(const DoubleInt &di) {
			return d > di.d;
		}
	};

	/**
	 * One state of the Viterbi algorithm.
	 */
	class ViterbiState {
	public:
		int tagIndex;
		int back;
		double best;
	};

	/**
	 * Used when counting n-best states for the current ViterbiState.
	 */
	class NBestState {
	public:
		int count;
		DoubleInt *states;

		NBestState(int n);
		~NBestState();
		DoubleInt *getState(int index);
	};

	Dictionary *dict;
	WordBuffer *wb;
	// size of history (n-gram), default is 3-gram
	int order;
	// used for saving tag history in processSentencePrune()
	int *prevTags;
	// how many states to keep for every word (0 = all)
	int prune;
	// structure where we keep top-scoring paths
	DoubleInt *pruneHeap;
	// how many states to record
	int nbest;
	// index of first item in the last state
	int lastFirst;
	// index of last item in the last state
	int lastLast;
	// common Viterbi states
	vector<ViterbiState> states;
	// n best pointers (double: score so far, int: index of previous state)
	vector<NBestState*> statesNBest;

	void setPrevProposedTags(int index, int count, int offset);
	void SaveStateNBest(int state, int back, double score);

	void processSentencePlain(Features *ftrs, int prevTagCount, bool useGamma);
	void Prune(int &currFirst, int &currLast);
	void processSentencePrune(Features *ftrs, int prevTagCount, bool useGamma);
public:
	/**
	 * Create a Viterbi object. Dictionary and WordBuffer is stored for future use.
	 * @param dict Dictionary.
	 * @param wb WordBuffer.
	 */
	Viterbi(Dictionary *dict, WordBuffer *wb);
	~Viterbi();

	/**
	 * Initialize a Viterbi object. Values are stored and used later
	 * while processing sentences.
	 * @param prune how many Viterbi states to keep per word. Lower scoring states are ignored.
	 * @param nbest how many top-scoring state alternatives to keep for one Viterbi state.
	 * @param order order of the model (usually 3).
	 */
	void init(int prune, int nbest, int order);
	/**
	 * Process one sentence and count best tag for every word.
	 * @param ftrs Features.
	 * @param prevTagCount how many tags before current words must be
	 * filled. This depends on the feature templates used.
	 */
	void processSentence(Features *ftrs, int prevTagCount, bool useGamma);
	/**
	 * Fill best sentence tags as proposed tags in the WordBuffer.
	 */
	void bestSentenceTags();
	/**
	 * Fill best sentence tags as proposed tags in the WordBuffer. This
	 * version uses n-best tags for every state. Score of every top-scoring
	 * tag assignment is filled.
	 * @param score double array to be filled with score of n top-scoring
	 * tag assignments. Must contain enough space (n elements), but may be
	 * NULL.
	 */
	void bestSentenceTagsNBest(double *score);

	/**
	 * Return string representation of Viterbi states. Result string should
	 * be deleted.
	 * @return String with Viterbi representation.
	 */
	string *toString();
};

inline Viterbi::NBestState::NBestState(int n) {
	count = 0;
	states = new DoubleInt[n];
}

inline Viterbi::NBestState::~NBestState() {
	delete[] states;
}

inline Viterbi::DoubleInt *Viterbi::NBestState::getState(int index) {
	assert(index < count);
	return &states[index];
}

#endif
