/**
 * @file
 * Single word representation. This class contain all attributes of a signle
 * word: features, tag, proposed tags, etc.
 */

#ifndef _WORD_H_
#define _WORD_H_

#include <config.h>

#include <assert.h>
#include <string>
#include "Dictionary.h"
#include "common.h"

using namespace std;

/**
 * Single word in a WordBuffer. Represents one line in the input file (word form, feature columns, tag, offer, ...).
 */
class Word {
	int *features;
	int correctTag;
	int additionalCorrectTag;
	int *proposedTagIndex;
	int proposedSize;
	int *offer;
	int offerSize;
	int *additionalOffer;

public:
	/**
	 * Constructs new Word with several constraints. The values are not
	 * stored (to reduce memory footprint).
	 * @param tagIndex index of the tag column.
	 * @param maxProposed Maximum number of proposed tags (used in nbest).
	 * @param maxOffer Maximum size of the offer.
	 */
	Word(int tagIndex, int maxProposed, int maxOffer);
	~Word();

	/**
	 * Parse a string and fill all attributes.
	 * @param dict Dictionary.
	 * @param appendDict append unknown strings to the dictionary.
	 * @param addCorrectTag append correct tag if it is not present in the offer (true in training).
	 * @param tagIndex index of the tag column.
	 * @param maxOffer Maximum size of the offer.
	 * @param str string to parse.
	 * @return False indicates parse error.
	 */
	bool parse(Dictionary *dict, bool appendDict, bool addCorrectTag, int tagIndex, int maxOffer, const char *str);
	/**
	 * Return value of a single feature.
	 * @param index index of a feature.
	 * @return dictionary-id of the value.
	 */
	int getFeatureValue(int index);
	/**
	 * Return correct tag value of a word.
	 * @return dictionary-id of the tag.
	 */
	int getCorrectTagValue();
	/**
	 * Return additional correct tag value of a word. Additional tag is
	 * separated from the tag itself by a space character. It may be used
	 * e.g. for the lemma in POS tagging.
	 * @return dictionary-id of the tag.
	 */
	int getAdditionalCorrectTagValue();
	/**
	 * Set index of a proposed tag.
	 * @param tagIndex index to a tag.
	 * @param index which proposed tag to work with.
	 */
	void setProposedTagIndex(int tagIndex, int index = 0);
	/**
	 * Return index of a proposed tag.
	 * @param index which proposed tag to work with.
	 * @return index of the tag.
	 */
	int getProposedTagIndex(int index = 0);
	/**
	 * Return proposed tag value.
	 * @param index which proposed tag to work with.
	 * @return dictionary-id of the tag.
	 */
	int getProposedTagValue(int index = 0);
	/**
	 * Return additional proposed tag value. \see getAdditionalCorrectTagValue
	 * @param index which proposed tag to work with.
	 * @return dictionary-id of the tag.
	 */
	int getAdditionalProposedTagValue(int index = 0);
	/**
	 * Return value of the offer.
	 * @param index which offer index to work with.
	 * @return dictionary-id of the offered tag.
	 */
	int getOfferValue(int index);
	/**
	 * Return additional value of the offer. \see getAdditionalCorrectTagValue
	 * @param index which offer index to work with.
	 * @return dictionary-id of the offered tag.
	 */
	int getAdditionalOfferValue(int index);
	/**
	 * Return number of tags in the offer.
	 * @return Size of the offer.
	 */
	int getOfferSize();
	/**
	 * Return whether current word does contain additional tags.
	 * @return True if current word does contain additional tags.
	 */
	bool hasAdditional();

	/**
	 * Return string representation of the Word. Result string should be
	 * deleted.
	 * @return String with word representation.
	 */
	string *toString(Dictionary *dict, int tagIndex);
};

inline int Word::getFeatureValue(int index) {
	return features[index];
}

inline int Word::getCorrectTagValue() {
	return correctTag;
}

inline int Word::getAdditionalCorrectTagValue() {
	return additionalCorrectTag;
}

inline void Word::setProposedTagIndex(int tagIndex, int index) {
	if (index > proposedSize)
		proposedSize = index+1;
	proposedTagIndex[index] = tagIndex;
}

inline int Word::getProposedTagIndex(int index) {
	return proposedTagIndex[index];
}

inline int Word::getProposedTagValue(int index) {
	return proposedTagIndex[index] == -1 ? DICTIONARY_NULL : getOfferValue(proposedTagIndex[index]);
}

inline int Word::getAdditionalProposedTagValue(int index) {
	return proposedTagIndex[index] == -1 ? DICTIONARY_NULL : getAdditionalOfferValue(proposedTagIndex[index]);
}

inline int Word::getOfferValue(int index) {
	assert(index >= 0 && index < offerSize);
	return offer[index];
}

inline int Word::getAdditionalOfferValue(int index) {
	assert(index >= 0 && index < offerSize);
	if (!additionalOffer)
		return DICTIONARY_NULL;
	return additionalOffer[index];
}

inline int Word::getOfferSize() {
	return offerSize;
}

inline bool Word::hasAdditional() {
	return additionalOffer != NULL;
}

#endif
