/**
 * WordBuffer
 *
 * Calling sequence:
 *
 * WordBuffer()		// create object
 * Init()		// read header
 */

#include <config.h>

#include <assert.h>
#include <string.h>
#include "WordBuffer.h"

WordBuffer::WordBuffer(Dictionary *dict) {
	this->dict = dict;
	this->appendDict = false;
	this->addCorrectTag = false;
	this->maxProposed = 1;
	this->maxOffer = 100;

	sentenceSize = -1;
}

WordBuffer::~WordBuffer() {
	for (vector<Word*>::iterator iter = words.begin(); iter != words.end(); ++iter) {
		delete *iter;
	}
	for (tr1::unordered_map<int, string*>::iterator iter = headerName.begin(); iter != headerName.end(); ++iter) {
		delete iter->second;
	}
}

void WordBuffer::init(bool appendDict, bool addCorrectTag, int maxProposed, int maxOffer) {
	this->appendDict = appendDict;
	this->addCorrectTag = addCorrectTag;
	this->maxProposed = maxProposed;
	this->maxOffer = maxOffer;

	for (vector<Word*>::iterator iter = words.begin(); iter != words.end(); ++iter) {
		delete *iter;
	}
	words.clear();

	sentenceSize = -1;
}

bool WordBuffer::initHeader(const char *dataHeader) {
	headerId.clear();
	headerName.clear();
	// create header hash
	char *str = strdup(dataHeader);
	chomp(str);
	char *helper = NULL;
	char *token = f_strtok_r(str, " \t", &helper);
	int i = 0;
	while (token) {
		headerId[token] = i;
		headerName[i] = new string(token);
		token = f_strtok_r(NULL, " \t", &helper);
		i++;
	}
	free(str);
	// last index is target (tag) index
	tagIndex = i-1;

	return true;
}

int WordBuffer::getHeaderIndex(const char *str) {
	tr1::unordered_map<string, int>::iterator iter = headerId.find(str);
	if (iter != headerId.end()) {
		return iter->second == tagIndex ? TAG_COLUMN : iter->second;
	}
	return ANY_COLUMN;
}

const char *WordBuffer::getHeaderName(int id) {
	if (id == TAG_COLUMN)
		id = tagIndex;
	tr1::unordered_map<int, string*>::iterator iter = headerName.find(id);
	if (iter != headerName.end()) {
		return iter->second->c_str();
	}
	return NULL;
}

bool WordBuffer::appendWord(const char *line, bool offerRequired) {
	// check that newSentence() was called
	assert(sentenceSize >= 0);
	// add new word to the current sentence
	if (sentenceSize < (int)words.size()) {
		if (!words[sentenceSize]->parse(dict, appendDict, addCorrectTag, tagIndex, maxOffer, line) && offerRequired)
			return false;
	} else {
		words.push_back(new Word(tagIndex, maxProposed, maxOffer));
		if (!words[sentenceSize]->parse(dict, appendDict, addCorrectTag, tagIndex, maxOffer, line) && offerRequired)
			return false;
	}
	sentenceSize++;
	return true;
}

void WordBuffer::newSentence() {
	sentenceSize = 0;
}

void WordBuffer::correctSentenceTags(vector<int> *v) {
	v->resize(sentenceSize);
	for (int i = 0; i < sentenceSize; i++) {
		(*v)[i] = getCorrectTagValue(i);
	}
}

int WordBuffer::evaluateSentence(bool compareAdditional) {
	int correct = 0;
	for (int i = 0; i < sentenceSize; i++) {
		int correctTag = getCorrectTagValue(i);
		for (int j = 0; j < maxProposed; j++) {
			int proposedTag = getProposedTagValue(i, j);
			if (proposedTag == DICTIONARY_NULL)
				break;
			if (proposedTag == correctTag) {
				if (!compareAdditional || getAdditionalProposedTagValue(i, j) == getAdditionalCorrectTagValue(i))
					correct++;
				break;
			}
		}
	}
	return correct;
}

string *WordBuffer::toString() {
	string *result = new string();

	result->append("WordBuffer:\n");
	for (int i = 0; i < sentenceSize; i++) {
		string *s = words[i]->toString(dict, tagIndex);
		result->append(*s);
		result->append("\n");
		delete s;
	}
	return result;
}
