/** @file */

#ifndef _WORD_BUFFER_H_
#define _WORD_BUFFER_H_

#include <config.h>

#include <string>
#include <vector>
#include <tr1/unordered_map>
#include "common.h"
#include "Dictionary.h"
#include "Word.h"

using namespace std;

/** Special mark for the tag feature column */
const int TAG_COLUMN = -1;
/** Special mark for any column */
const int ANY_COLUMN = -2;

/**
 * WordBuffer representes one sentence. It consists of a vector of Word
 * elements that grows as needed. Currently, it is not shrinked after the
 * sentence was processed.
 */

class WordBuffer {
	Dictionary *dict;
	bool appendDict;	// whether to add unseen word into the dictionary
	bool addCorrectTag;
	int maxProposed;
	int maxOffer;

	// header fields translation hash
	tr1::unordered_map<string, int> headerId;
	tr1::unordered_map<int, string*> headerName;
	int tagIndex;		// index of target tag

	vector<Word*> words;	// one sentence, gets larger and larger; may be truncated from time to time
	int sentenceSize;	// really used items of words vector
public:
	/**
	 * Create a WordBuffer. Dictionary is stored for future use.
	 * @param dict Dictionary.
	 */
	WordBuffer(Dictionary *dict);
	~WordBuffer();

	/**
	 * (Re-)initialize a WordBuffer. Values are stored for futher use.
	 * @param appendDict true when unknown words should be added in the dictionary.
	 * @param addCorrectTag append correct tag if it is not present in the offer (true in training).
	 * @param maxProposed Maximum number of proposed tags (used in nbest).
	 * @param maxOffer Maximum size of the offer.
	 */
	void init(bool appendDict, bool addCorrectTag, int maxProposed, int maxOffer);
	/**
	 * (Re-)initialize columns and column names of a WordBuffer. Column
	 * names are used to describe features and are separated by tabs. Last
	 * column name is a tag.
	 * @param dataHeader string header.
	 * @return Whether parsing of the header was successful. Currently it
	 * is always successful.
	 */
	bool initHeader(const char *dataHeader);
	/**
	 * Return index of a column.
	 * @param str name of the column.
	 * @return id of the column.
	 */
	int getHeaderIndex(const char *str);
	/**
	 * Return id of a column name.
	 * @param id column id.
	 * @return String with the name of the column.
	 */
	const char *getHeaderName(int id);
	/**
	 * Append new word to the WordBuffer. Word is represented by all
	 * features (passed as a string where feature values are strings
	 * separated by tabs). Correct tags and offered tags are appended after
	 * the feature values.
	 * @param line string representation of features, etc.
	 * @param offerRequired whether there must be at least one tag in the offer.
	 * @return True if word was parsed correctly.
	 */
	bool appendWord(const char *line, bool offerRequired);
	/**
	 * Reset the current state of the WordBuffer, so that new sentence can be processed.
	 */
	void newSentence();
	/**
	 * Return number of feature columns (+tag).
	 * @return Number of columns.
	 */
	int getHeaderCount();
	/**
	 * Return value of a single feature.
	 * @param index index of a feature.
	 * @param row index of the word in the sentence.
	 * @return dictionary-id of the value.
	 */
	int getFeatureValue(int row, int index);
	/**
	 * Return correct tag value of a word.
	 * @param row index of the word in the sentence.
	 * @return dictionary-id of the tag.
	 */
	int getCorrectTagValue(int row);
	/**
	 * Return additional correct tag value of a word. Additional tag is
	 * separated from the tag itself by a space character. It may be used
	 * e.g. for the lemma in POS tagging.
	 * @param row index of the word in the sentence.
	 * @return dictionary-id of the tag.
	 */
	int getAdditionalCorrectTagValue(int row);
	/**
	 * Set index of a proposed tag.
	 * @param row index of the word in the sentence.
	 * @param tagIndex index to a tag.
	 * @param index which proposed tag to work with.
	 */
	void setProposedTagIndex(int row, int tagIndex, int index = 0);
	/**
	 * Return proposed tag value.
	 * @param row index of the word in the sentence.
	 * @param index which proposed tag to work with.
	 * @return dictionary-id of the tag.
	 */
	int getProposedTagIndex(int row, int index = 0);
	/**
	 * Return value of the offer.
	 * @param row index of the word in the sentence.
	 * @param index which offer index to work with.
	 * @return dictionary-id of the offered tag.
	 */
	int getProposedTagValue(int row, int index = 0);
	/**
	 * Return additional proposed tag value. \see getAdditionalCorrectTagValue
	 * @param row index of the word in the sentence.
	 * @param index which proposed tag to work with.
	 * @return dictionary-id of the tag.
	 */
	int getAdditionalProposedTagValue(int row, int index = 0);
	/**
	 * Return additional value of the offer. \see getAdditionalCorrectTagValue
	 * @param row index of the word in the sentence.
	 * @param index which offer index to work with.
	 * @return dictionary-id of the offered tag.
	 */
	int getOfferValue(int row, int index);
	/**
	 * Return additional value of the offer. \see getAdditionalCorrectTagValue
	 * @param row index of the word in the sentence.
	 * @param index which offer index to work with.
	 * @return dictionary-id of the offered tag.
	 */
	int getAdditionalOfferValue(int row, int index);
	/**
	 * Return number of tags in the offer.
	 * @param row index of the word in the sentence.
	 * @return Size of the offer.
	 */
	int getOfferSize(int row);
	/**
	 * Return size of a sentence currently in the word buffer.
	 * @return Size of a sentence in words.
	 */
	int getSentenceSize();
	/**
	 * Return whether current word does contain additional tags.
	 * @param row index of the word in the sentence.
	 * @return True if current word does contain additional tags.
	 */
	bool hasAdditional(int row);
	/**
	 * Fill vector with correct tags. Vector size is adjusted according to
	 * the sentence size.
	 * @param v vector to be filled.
	 */
	void correctSentenceTags(vector<int> *v);
	/**
	 * Return number of tags that were correcty assigned in the current sentence.
	 * @param compareAdditional also take into account additional tags (aka lemmas).
	 * @return Number of correct tags.
	 */
	int evaluateSentence(bool compareAdditional = false);

	/**
	 * Return string representation of the WordBuffer. Result string should
	 * be deleted.
	 * @return String with WordBuffer representation.
	 */
	string *toString();
};

inline int WordBuffer::getHeaderCount() {
	return headerName.size();
}

inline int WordBuffer::getFeatureValue(int row, int index) {
	return (row >= 0 && row < sentenceSize) ? words[row]->getFeatureValue(index) : DICTIONARY_NULL;
}

inline int WordBuffer::getCorrectTagValue(int row) {
	return (row >= 0 && row < sentenceSize) ? words[row]->getCorrectTagValue() : DICTIONARY_NULL;
}

inline int WordBuffer::getAdditionalCorrectTagValue(int row) {
	return (row >= 0 && row < sentenceSize) ? words[row]->getAdditionalCorrectTagValue() : DICTIONARY_NULL;
}

inline void WordBuffer::setProposedTagIndex(int row, int tagIndex, int index) {
	assert(index < maxProposed);
	if (row >= 0 && row < sentenceSize)
		words[row]->setProposedTagIndex(tagIndex, index);
}

inline int WordBuffer::getProposedTagIndex(int row, int index) {
	assert(index < maxProposed);
	return (row >= 0 && row < sentenceSize) ? words[row]->getProposedTagIndex(index) : DICTIONARY_NULL;
}

inline int WordBuffer::getProposedTagValue(int row, int index) {
	assert(index < maxProposed);
	return (row >= 0 && row < sentenceSize) ? words[row]->getProposedTagValue(index) : DICTIONARY_NULL;
}

inline int WordBuffer::getAdditionalProposedTagValue(int row, int index) {
	assert(index < maxProposed);
	return (row >= 0 && row < sentenceSize) ? words[row]->getAdditionalProposedTagValue(index) : DICTIONARY_NULL;
}

inline int WordBuffer::getOfferValue(int row, int index) {
	return (row >= 0 && row < sentenceSize) ? words[row]->getOfferValue(index) : DICTIONARY_NULL;
}

inline int WordBuffer::getAdditionalOfferValue(int row, int index) {
	return (row >= 0 && row < sentenceSize) ? words[row]->getAdditionalOfferValue(index) : DICTIONARY_NULL;
}

inline int WordBuffer::getOfferSize(int row) {
	return (row >= 0 && row < sentenceSize) ? words[row]->getOfferSize() : -1;
}

inline int WordBuffer::getSentenceSize() {
	return sentenceSize;
}

inline bool WordBuffer::hasAdditional(int row) {
	return (row >= 0 && row < sentenceSize) ? words[row]->hasAdditional() : false;
}

#endif
