/**
 * Common functions, macros, etc.
 */

#ifndef _COMMON_H_
#define _COMMON_H_

#include <config.h>

#include <ctype.h>
#include <errno.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <string>
#include <tr1/unordered_set>

using namespace std;

/** Maximum size of a line, that can be parsed */
const int MAX_LINE_SIZE = 100*1024;

/**
 * Print a message and exit. If errno != 0, print also strerror. Uses stderr.
 * @param msg message to print before strerror string.
 */
inline void die(const char *msg, ...) {
	va_list args;
	va_start(args, msg);
	vfprintf(stderr, msg, args);
	va_end(args);
	if (errno != 0)
		fprintf(stderr, ": %s", strerror(errno));
	fputc('\n', stderr);
	exit(1);
}

/**
 * Print an error message. If errno != 0, print also strerror. Uses stderr.
 * @param msg message to print before strerror string.
 */
inline void error(const char *msg, ...) {
	va_list args;
	va_start(args, msg);
	vfprintf(stderr, msg, args);
	va_end(args);
	if (errno != 0)
		fprintf(stderr, ": %s", strerror(errno));
	fputc('\n', stderr);
}

/**
 * Swap endianness of a 32-bit integer.
 */
inline int32_t endian_swap(int32_t x) {
	return (x>>24) | 
	       ((x<<8) & 0x00FF0000) |
	       ((x>>8) & 0x0000FF00) |
	       (x<<24);
}

/**
 * Swap endianness of a 32-bit unsigned.
 */
inline uint32_t endian_swap(uint32_t x) {
	return (x>>24) | 
	       ((x<<8) & 0x00FF0000) |
	       ((x>>8) & 0x0000FF00) |
	       (x<<24);
}

/**
 * Swap endianness of a 64-bit integer.
 */
inline int64_t endian_swap(int64_t x) {
	return (x>>56) | 
	       ((x<<40) & 0x00FF000000000000LL) |
	       ((x<<24) & 0x0000FF0000000000LL) |
	       ((x<<8)  & 0x000000FF00000000LL) |
	       ((x>>8)  & 0x00000000FF000000LL) |
	       ((x>>24) & 0x0000000000FF0000LL) |
	       ((x>>40) & 0x000000000000FF00LL) |
	       (x<<56);
}

/**
 * Return whether string is a comment. Comment line contains spaces followed by a hash character (#).
 * @param s string to check.
 * @return True if the line is a comment.
 */
bool isComment(const char *s);

/**
 * Return whether string is empty. White-spaces are ignored.
 * @param s string to check.
 * @return True if string contains only white-spaces.
 */
bool isEmpty(const char *s);

/**
 * Return Number of character instances in a string.
 * @param s string to check.
 * @param c character to count.
 * @return Number of instances of c in s.
 */
int countChars(const char *s, char c);

/**
 * Removes leading white-spaces.
 * @param s string to check.
 * @return First non-white-space character.
 */
char *skipWhiteSpace(char *s);

/**
 * Convert string to an integer. Using decimal notation, plus and minus signs are allowed.
 * @param str string to parse, updated to point after last character of the integer.
 * @return Result integer, defaults to 0.
 */
int readInt(char **str);

/**
 * Remove white-spaces from the end of the string. Similar to Perl's chomp function.
 * @param s string to update.
 */
void chomp(char *s);

/**
 * strtok_r implementation
 */
inline char *f_strtok_r(char *str, const char *delim, char **nextp) {
	char *ret;

	if (str == NULL)
		str = *nextp;
	str += strspn(str, delim);
	if (*str == '\0')
		return NULL;
	ret = str;
	str += strcspn(str, delim);
	if (*str)
		*str++ = '\0';
	*nextp = str;
	return ret;
}

#define SKIPWS(s) s = skipWhiteSpace(s)

#define PRINT_STRING(s) { string *x = s; fputs(s->c_str(), stderr); delete x; }
#define PRINT_NL() fputc('\n', stderr);

#endif
