/* C interface implementation */

#include <config.h>

#include <getopt.h>
#include <stdio.h>
#include <limits>
#include <string.h>
#include "common.h"
#include "featurama/perc.h"
#include "Dictionary.h"
#include "WordBuffer.h"
#include "Features.h"
#include "Viterbi.h"

enum perc_type {
	PERC_NULL = 0x01,
	PERC_MAKE_FEATURES = 0x02,
	PERC_TRAIN = 0x04,
	PERC_TEST = 0x08
};

struct perc {
	enum perc_type type;
	Dictionary *dict;
	WordBuffer *wb;
	Features *ftrs;
	Viterbi *viterbi;
	int minTag;
	int nbest;
	double *score;
	int wordCount;
	int correctCount;
	int additionalCorrectCount;
};

static inline bool _check_perc_type(PERC *perc, int mask) {
	if (perc->type & mask)
		return true;
	errno = 0;
	error("Invalid perc handle (not initialized)");
	return false;
}

PERC *perc_init() {
	PERC *perc = new PERC();
	perc->type = PERC_NULL;
	perc->dict = NULL;
	perc->wb = NULL;
	perc->ftrs = NULL;
	perc->viterbi = NULL;
	perc->minTag = 0;
	perc->nbest = 1;
	perc->score = NULL;
	perc->wordCount = 0;
	perc->correctCount = 0;
	perc->additionalCorrectCount = 0;

	return perc;
}

void perc_destroy(PERC *perc) {
	delete perc->viterbi;
	delete perc->ftrs;
	delete perc->wb;
	delete perc->dict;
	delete[] perc->score;
	delete perc;
}

int perc_process_file_full(PERC *perc, const char *dataFile, int skipHeader) {
	int result;

	if (!_check_perc_type(perc, PERC_MAKE_FEATURES|PERC_TRAIN|PERC_TEST))
		return 0;

	FILE *f = fopen(dataFile, "r");
	if (!f) {
		error("Cannot open file (%s)", dataFile);
		return 0;
	}
	if (skipHeader) {
		// ignore header line
		char str[MAX_LINE_SIZE];
		errno = 0;
		if (!fgets(str, sizeof(str), f)) {
			error("Cannot read data header (%s)", dataFile);
			return 0;
		}
	}
	result = perc_process_file(perc, f);
	fclose(f);

	return result;
}

int perc_process_file(PERC *perc, FILE *dataFile) {
	char str[MAX_LINE_SIZE];
	int eof = 0;
	int eos = 0;

	if (!_check_perc_type(perc, PERC_MAKE_FEATURES|PERC_TRAIN|PERC_TEST))
		return 0;

	perc_begin_sentence(perc);
	do {
		errno = 0;
		if (!fgets(str, sizeof(str), dataFile)) {
			if (errno > 0) {
				error("Cannot read from file: ");
				return 0;
			}
			eof = 1;
		} else {
			chomp(str);
			if (!isEmpty(str)) {
				if (eos) {
					perc_end_sentence(perc);
					perc_begin_sentence(perc);
					eos = 0;
				}
				if (!perc->wb->appendWord(str, perc->type & (PERC_TRAIN|PERC_TEST))) {
					error("Too few columns (word %d: %s)", perc->wordCount+1, str);
				} else {
					perc->wordCount++;
				}
			} else {
				eos = 1;
			}
		}
	} while (!eof);
	perc_end_sentence(perc);

	return 1;
}

int perc_append_word(PERC *perc, const char *line) {
	if (!_check_perc_type(perc, PERC_MAKE_FEATURES|PERC_TRAIN|PERC_TEST))
		return 0;
	if (!perc->wb->appendWord(line, perc->type & (PERC_TRAIN|PERC_TEST))) {
		error("Too few columns (word %d: %s)", perc->wordCount+1, line);
		return 0;
	} else {
		perc->wordCount++;
	}
	return 1;
}

void perc_begin_sentence(PERC *perc) {
	if (!_check_perc_type(perc, PERC_MAKE_FEATURES|PERC_TRAIN|PERC_TEST))
		return;
	perc->wb->newSentence();
}

void perc_end_sentence(PERC *perc) {
	if (!_check_perc_type(perc, PERC_MAKE_FEATURES|PERC_TRAIN|PERC_TEST))
		return;

	int sentenceSize = perc->wb->getSentenceSize();
	if (!sentenceSize)
		return;

	switch (perc->type) {
	case PERC_MAKE_FEATURES:
		for (int i = 0; i < sentenceSize; i++)
			perc->ftrs->expandFeatureTemplates(i);
		break;
	case PERC_TRAIN:
		perc->viterbi->processSentence(perc->ftrs, perc->minTag == std::numeric_limits<int>::min() ? sentenceSize : -perc->minTag, false);
		perc->viterbi->bestSentenceTags();
		perc->ftrs->updateAlphas();

		perc->correctCount += perc->wb->evaluateSentence();
		if (perc->wb->hasAdditional(0))
			perc->additionalCorrectCount += perc->wb->evaluateSentence(true);
		break;
	case PERC_TEST:
		perc->viterbi->processSentence(perc->ftrs, perc->minTag == std::numeric_limits<int>::min() ? sentenceSize : -perc->minTag, true);
		if (perc->nbest > 1)
			perc->viterbi->bestSentenceTagsNBest(perc->score);
		else
			perc->viterbi->bestSentenceTags();

		perc->correctCount += perc->wb->evaluateSentence();
		if (perc->wb->hasAdditional(0))
			perc->additionalCorrectCount += perc->wb->evaluateSentence(true);
		break;
	default:
		break;
	}
}

void perc_get_accuracy(PERC *perc, double *correct, double *additionalCorrect) {
	if (!_check_perc_type(perc, PERC_TRAIN|PERC_TEST))
		return;

	if (correct)
		*correct = perc->wordCount > 0 ? (double)perc->correctCount / perc->wordCount : 0;
	if (additionalCorrect)
		*additionalCorrect = perc->wb->hasAdditional(0) ? (perc->wordCount > 0 ? (double)perc->additionalCorrectCount / perc->wordCount : 0) : -1;
}

void perc_reset_accuracy(PERC *perc) {
	if (!_check_perc_type(perc, PERC_TRAIN|PERC_TEST))
		return;
	perc->wordCount = 0;
	perc->correctCount = 0;
	perc->additionalCorrectCount = 0;
}

const char *perc_get_proposed_tag(PERC *perc, int sentenceIndex, int index) {
	if (!_check_perc_type(perc, PERC_TEST))
		return NULL;

	return perc->dict->getString(perc->wb->getProposedTagValue(sentenceIndex, index));
}

const char *perc_get_additional_proposed_tag(PERC *perc, int sentenceIndex, int index) {
	if (!_check_perc_type(perc, PERC_TEST))
		return NULL;

	int additionalTag = perc->wb->getAdditionalProposedTagValue(sentenceIndex, index);
	return additionalTag != DICTIONARY_NULL ? perc->dict->getString(additionalTag) : NULL;
}

double perc_get_nbest_score(PERC *perc, int index) {
	if (!_check_perc_type(perc, PERC_TEST))
		return 0.0;

	if (!perc->score) {
		errno = 0;
		error("No nbest specified");
		return -1.0;
	}

	if (index < 0 || index >= perc->nbest) {
		errno = 0;
		error("Invalid score index %d", index);
		return -1.0;
	}
	return perc->score[index];
}

int perc_make_features_init(PERC *perc, const char *templateFile, const char *dictFile, const char *header) {
	if (!perc->dict)
		perc->dict = new Dictionary();
	if (dictFile) {
		if (!perc->dict->load(dictFile)) {
			error("Cannot load dictionary (%s)", dictFile);
			return 0;
		}
	}

	if (!perc->wb)
		perc->wb = new WordBuffer(perc->dict);
	perc->wb->init(true, false, 1, 5000);
	if (header) {
		if (!perc->wb->initHeader(header)) {
			error("Invalid header, cannot init word buffer");
			return 0;
		}
	}

	if (!perc->ftrs)
		perc->ftrs = new Features(perc->dict, perc->wb);
	else
		perc->ftrs->init();
	if (!templateFile) {
		errno = 0;
		error("Feature template must be specified");
		return 0;
	}
	if (!perc->ftrs->loadFeatureTemplates(templateFile)) {
		error("Cannot load feature templates (%s)", templateFile);
		return 0;
	}

	perc->type = PERC_MAKE_FEATURES;

	return 1;
}

int perc_make_features_finish(PERC *perc, const char *featureFile, const char *dictFile, int minFreq) {
        if (featureFile && !perc->ftrs->saveExpandedFeatures(featureFile, minFreq)) {
		error("Cannot save features (%s)", featureFile);
		return 0;
	}
        if (dictFile && !perc->dict->save(dictFile)) {
		error("Cannot save dictionary (%s)", dictFile);
		return 0;
	}

	perc->type = PERC_NULL;
	perc->wordCount = 0;

	return 1;
}

int perc_train_init(PERC *perc, const char *featureFile, const char *dictFile, const char *alphaFile, const char *header, int prune, int order) {
	if (!perc->dict)
		perc->dict = new Dictionary();
	if (dictFile) {
		if (!perc->dict->load(dictFile)) {
			error("Cannot load dictionary (%s)", dictFile);
			return 0;
		}
	}

	if (!perc->wb)
		perc->wb = new WordBuffer(perc->dict);
	perc->wb->init(true, true, 1, 5000);
	if (header) {
		if (!perc->wb->initHeader(header)) {
			error("Invalid header, cannot init word buffer");
			return 0;
		}
	}

	if (!perc->ftrs)
		perc->ftrs = new Features(perc->dict, perc->wb);
	if (featureFile && !perc->ftrs->loadFeatures(featureFile)) {
		error("Cannot load feature file (%s)", featureFile);
		return 0;
	}
	int maxTag;
	perc->ftrs->getTagContextSize(&perc->minTag, &maxTag);
	if (maxTag != 0) {
		errno = 0;
		error("Invalid feature file (maxTag = %d)", maxTag);
		return 0;
	}

	if (alphaFile) {
		if (!perc->ftrs->loadAlphas(alphaFile)) {
			error("Cannot load alphas (%s)", alphaFile);
			return 0;
		}
	} else {
		perc->ftrs->initAlphas();
	}

	if (!perc->viterbi)
		perc->viterbi = new Viterbi(perc->dict, perc->wb);
	perc->viterbi->init(prune, 0, order);

	perc->type = PERC_TRAIN;
	perc->wordCount = 0;
	perc->correctCount = 0;
	perc->additionalCorrectCount = 0;

	return 1;
}

int perc_train_finish(PERC *perc, const char *dictFile, const char *alphaFile) {
	perc->ftrs->finishAlphas();
        if (alphaFile && !perc->ftrs->saveAlphas(alphaFile)) {
		error("Cannot save alphas (%s)", alphaFile);
		return 0;
	}

        if (dictFile && !perc->dict->save(dictFile)) {
		error("Cannot save dictionary (%s)", dictFile);
		return 0;
	}

	perc->type = PERC_NULL;

	return 1;
}

int perc_test_init(PERC *perc, const char *featureFile, const char *dictFile, const char *alphaFile, const char *header, int prune, int nbest, int order) {
	if (nbest <= 1)
		nbest = 1;
	perc->nbest = nbest;
	delete[] perc->score;
	perc->score = nbest > 1 ? new double[nbest] : NULL;

	if (!perc->dict)
		perc->dict = new Dictionary();
	if (dictFile) {
		if (!perc->dict->load(dictFile)) {
			error("Cannot load dictionary (%s)", dictFile);
			return 0;
		}
	}

	if (!perc->wb)
		perc->wb = new WordBuffer(perc->dict);
	perc->wb->init(true, false, nbest, 5000);
	if (header) {
		if (!perc->wb->initHeader(header)) {
			error("Invalid header, cannot init word buffer");
			return 0;
		}
	}

	if (!perc->ftrs)
		perc->ftrs = new Features(perc->dict, perc->wb);
	if (featureFile && !perc->ftrs->loadFeatures(featureFile)) {
		error("Cannot load feature file (%s)", featureFile);
		return 0;
	}
	int maxTag;
	perc->ftrs->getTagContextSize(&perc->minTag, &maxTag);
	if (maxTag != 0) {
		errno = 0;
		error("Invalid feature file (maxTag = %d)", maxTag);
		return 0;
	}

	if (alphaFile) {
		if (!perc->ftrs->loadAlphas(alphaFile)) {
			error("Cannot load alphas (%s)", alphaFile);
			return 0;
		}
	} else {
		perc->ftrs->initAlphas();
	}

	if (!perc->viterbi)
		perc->viterbi = new Viterbi(perc->dict, perc->wb);
	perc->viterbi->init(prune, nbest, order);

	perc->type = PERC_TEST;
	perc->wordCount = 0;
	perc->correctCount = 0;
	perc->additionalCorrectCount = 0;

	return 1;
}

int perc_test_finish(PERC *perc) {
	delete[] perc->score;
	perc->score = NULL;

	perc->type = PERC_NULL;

	return 1;
}
