#!/usr/bin/env perl

use strict;
use warnings;
use Featurama::Perc;
use Getopt::Long;
Getopt::Long::Configure('no_ignore_case');
use POSIX qw(strftime floor);

BEGIN {
	binmode STDIN, ":utf8";
	binmode STDOUT, ":utf8";
	binmode STDERR, ":utf8";
}

my $command = 'unknown';
my $feature_templates;
my $input_features;
my $output_features;
my $input_dictionary;
my $output_dictionary;
my $input_alphas;
my $output_alphas;
my $min_freq = 1;
my $iterations = 1;
my $prune = 0;
my $nbest = 1;
my $nbest_score = 0;
my $order = 3;

my $verbose = 0;
my $help = 0;
my $version = 0;

sub printHelp() {
	if ($command eq 'make_features') {
		print <<END
usage: perc.pl make_features [options] data_file ...
options:
  --feature-templates, -t	Feature templates (default.ft)
  --output-features, -F		Features (default.f)
  --input-dictionary, -d	Input dictionary (none)
  --output-dictionary, -D	Output dictionary (default.dict)
  --min-freq, -n		Minimum feature frequency (1)
  --verbose, -v			Be verbose
  --help, -h			This help
  --version, -V			Version information
END
	} elsif ($command eq 'train') {
		print <<END
usage: perc train [options] data_file ...
options:
  --input-features, -f		Features file (deafult.f)
  --input-dictionary, -d	Input dictionary file (default.dict)
  --output-dictionary, -D	Output dictionary file (none)
  --input-alphas, -a		Input alphas file (none)
  --output-alphas, -A		Output alphas file (default.alpha)
  --iterations, -i		Number of iterations (1)
  --prune, -p			Prune: limit number of states (keep all)
  --order, -o			Model order (n-gram) (3)
  --verbose, -v			Be verbose
  --help, -h			This help
  --version, -V			Version information
END
	} elsif ($command eq 'test') {
		print <<END
usage: perc test [options] data_file ...
options:
  --input-features, -f		Features file (default.f)
  --input-dictionary, -d	Dictionary file (default.dict)
  --input-alphas, -a		Alphas file (default.alpha)
  --prune, -p			Prune: limit number of states (keep all)
  --nbest, -b			NBest: output n top-scoring solutions (1)
  --nbest-score, -s		NBest: show score for proposed solution (no)
  --order, -o			Model order (n-gram) (3)
  --verbose, -v			Be verbose
  --help, -h			This help
  --version, -V			Version information
END
	} else {
		print <<END
usage: perc <make_features|train|test> [options] data_file ...
Use perc [command] --help to get info about options.
END
	}
	exit(0);
}

sub printVersion() {
	print "perc.pl (??bit alpha), %s(%s)\n";
	exit(0);
}

sub getOptions() {
	GetOptions(	"feature-templates|t=s" => \$feature_templates,
			"input-dictionary|d=s" => \$input_dictionary,
			"output-dictionary|D=s" => \$output_dictionary,
			"input-features|f=s" => \$input_features,
			"output-features|F=s" => \$output_features,
			"input-alphas|a=s" => \$input_alphas,
			"output-alphas|A=s" => \$output_alphas,
			"min-freq|n=i" => \$min_freq,
			"iterations|i=i" => \$iterations,
			"prune|p=i" => \$prune,
			"order|o=i" => \$order,
			"nbest|b=i" => \$nbest,
			"nbest-score|s" => \$nbest_score,
			"verbose|v" => \$verbose,
			"help|h" => \$help,
			"version|V" => \$version) or die;
}

sub main() {
	&getOptions();
	&printHelp() if (@ARGV == 0);
	$command = shift(@ARGV);
	if ($command eq 'make_features') {
		$feature_templates = 'default.ft' if (not defined $feature_templates);
		$output_features = 'default.f' if (not defined $output_features);
		$output_dictionary = 'default.dict' if (not defined $output_dictionary);
	} elsif ($command eq 'train') {
		$input_features = 'default.f' if (not defined $input_features);
		$input_dictionary = 'default.dict' if (not defined $input_dictionary);
		$output_alphas = 'default.alpha' if (not defined $output_alphas);
	} elsif ($command eq 'test') {
		$input_features = 'default.f' if (not defined $input_features);
		$input_dictionary = 'default.dict' if (not defined $input_dictionary);
		$input_alphas = 'default.alpha' if (not defined $input_alphas);
	}
	&printHelp() if ($help || $command eq 'undefined');
	&printHelp() if (@ARGV == 0);

	my $fin;
	open($fin, '<'.$ARGV[0]) or die("Cannot open file ".$ARGV[0]);
	my $header = <$fin>;

	my $perc = Featurama::Perc->new();
	if ($command eq 'make_features') {
		$perc->makeFeaturesInit($feature_templates, $input_dictionary, $header) or die;
	} elsif ($command eq 'train') {
		$perc->trainInit($input_features, $input_dictionary, $input_alphas, $header, $prune, $order) or die;
	} elsif ($command eq 'test') {
		$perc->testInit($input_features, $input_dictionary, $input_alphas, $header, $prune, $nbest, $order) or die;
	}

	print STDERR "0 kw processed\n" if ($verbose);

	my $fileIndex = 0;

	for (my $iter = 0; $iter < $iterations; $iter++) {
		my $total = 0;
		my $last = 0;
		my $firstSentence = 1;

		while ($fileIndex < @ARGV) {
			if (not defined $fin) {
				open($fin, '<'.$ARGV[$fileIndex]) or die("Cannot open file ".$ARGV[$fileIndex]);
				$header = <$fin>;
			}
	
			my $eof = 0;
			my $eos = 0;
			my $sentenceSize = 0;
			$perc->beginSentence();
			do {
				my $str = <$fin>;
				if (not defined $str) {
					$eof = 1;
				} else {
					chomp($str);
				}
				if ($eof or $str ne '') {
					if ($eos or $eof) {
						$perc->endSentence();
						if ($command eq 'make_features') {
							if ($verbose) {
								printf STDERR "%d kw processed\n", POSIX::floor($total/1000) if (POSIX::floor($last/10000) < POSIX::floor($total/10000));
								$last = $total;
							}
						} elsif ($command eq 'train') {
							if ($verbose) {
								if (POSIX::floor($last/10000) < POSIX::floor($total/10000)) {
									my ($c, $ac) = $perc->getAccuracy();
									printf STDERR "%d kw processed, accuracy: %.3f", POSIX::floor($total/1000), 100*$c;
									printf STDERR " (%.3f)", 100*$ac if ($ac >= 0);
									print STDERR "\n";
								}
								$last = $total;
							}
						} elsif ($command eq 'test') {
							if ($firstSentence) {
								my $now = strftime("%a %b %e %H:%M:%S %Y", localtime);
								printf "# Output%s created: %s\n", $nbest_score ? " (score)" : "", $now;
								$firstSentence = 0;
							} else {
								print "\n";
							}

							if ($nbest_score) {
								for (my $i = 0; $i < $nbest; $i++) {
									print '\t' if ($i > 0);
									printf "%.3lf", $perc->getScore($i);
								}
								print "\n";
							}
							for (my $i = 0; $i < $sentenceSize; $i++) {
								for (my $j = 0; $j < $nbest; $j++) {
									print '\t' if ($j > 0);
									print $perc->getProposedTag($i, $j);
									my $s = $perc->getAdditionalProposedTag($i, $j);
									print ' '.$s if (defined $s);
								}
								print "\n";
							}

							if ($verbose) {
								if (POSIX::floor($last/10000) < POSIX::floor($total/10000)) {
									my ($c, $ac) = $perc->getAccuracy();
									printf STDERR "%d kw processed, accuracy: %.3f", POSIX::floor($total/1000), 100*$c;
									printf STDERR " (%.3f)", 100*$ac if ($ac >= 0);
									print STDERR "\n";
								}
								$last = $total;
							}
						}
						$eos = 0;
						$sentenceSize = 0;
						$perc->beginSentence() if (!$eof);
					}
					if (!$eof) {
						if ($perc->appendWord($str)) {
							$total++;
							$sentenceSize++;
						}
					}
				} else {
					$eos = 1;
				}
			} while (!eof);
	
			close($fin);
			$fin = undef;
			$fileIndex++;
		}

		if ($command eq 'make_features') {
			printf STDERR "Words processed: %d\n", $total if ($verbose);
		} elsif ($command eq 'train') {
			if ($verbose) {
				my ($c, $ac) = $perc->getAccuracy();
				$perc->resetAccuracy();
				printf STDERR "Iteration: %d, words processed: %d, accuracy: %.3f", $iter+1, $total, 100*$c;
				printf STDERR " (%.3f)", 100*$ac if ($ac >= 0);
				print STDERR "\n";
			}
		} elsif ($command eq 'test') {
			my ($c, $ac) = $perc->getAccuracy();
			printf STDERR "%d words processed, accuracy: %.3f", $total, 100*$c;
			printf STDERR " (%.3f)", 100*$ac if ($ac >= 0);
			print STDERR "\n";
		}

		$fileIndex = 0;
	}

	if ($command eq 'make_features') {
		$perc->makeFeaturesFinish($output_features, $output_dictionary, $min_freq) or die;
	} elsif ($command eq 'train') {
		$perc->trainFinish($output_dictionary, $output_alphas) or die;
	} elsif ($command eq 'test') {
		$perc->testFinish() or die();
	}

	return 0;
}

&main();
