/**
 * Make features && train && test
 */

#include <config.h>

#include <getopt.h>
#include <stdio.h>
#include <string.h>
#include "common.h"
#include "featurama/perc.h"
#include "version.h"

typedef enum {
	CMD_NULL = 0,
	CMD_MAKE_FEATURES = 1,
	CMD_TRAIN = 2,
	CMD_TEST = 3,
} cmd_t;

cmd_t command = CMD_NULL;
char *feature_templates = NULL;
char *input_features = NULL;
char *output_features = NULL;
char *input_dictionary = NULL;
char *output_dictionary = NULL;
char *input_alphas = NULL;
char *output_alphas = NULL;
int min_freq = 1;
int iterations = 1;
int prune = 0;
int nbest = 1;
bool nbest_score = false;
int order = 3;

int verbose = 0;
bool help = false;

void print_help() {
	switch (command) {
	case CMD_MAKE_FEATURES:
		fprintf(stderr, "\
usage: perc make_features [options] data_file ...\n\
options:\n\
  --feature-templates, -t\tFeature templates (default.ft)\n\
  --output-features, -F\t\tFeatures (default.f)\n\
  --input-dictionary, -d\tInput dictionary (none)\n\
  --output-dictionary, -D\tOutput dictionary (default.dict)\n\
  --min-freq, -n\t\tMinimum feature frequency (1)\n\
  --verbose, -v\t\t\tBe verbose\n\
  --help, -h\t\t\tThis help\n\
  --version, -V\t\t\tVersion information\n");
		break;
	case CMD_TRAIN:
		fprintf(stderr, "\
usage: perc train [options] data_file ...\n\
options:\n\
  --input-features, -f\t\tFeatures file (deafult.f)\n\
  --input-dictionary, -d\tInput dictionary file (default.dict)\n\
  --output-dictionary, -D\tOutput dictionary file (none)\n\
  --input-alphas, -a\t\tInput alphas file (none)\n\
  --output-alphas, -A\t\tOutput alphas file (default.alpha)\n\
  --iterations, -i\t\tNumber of iterations (1)\n\
  --prune, -p\t\t\tPrune: limit number of states (keep all)\n\
  --order, -o\t\t\tModel order (n-gram) (3)\n\
  --verbose, -v\t\t\tBe verbose\n\
  --help, -h\t\t\tThis help\n\
  --version, -V\t\t\tVersion information\n");
		break;
	case CMD_TEST:
		fprintf(stderr, "\
usage: perc test [options] data_file ...\n\
options:\n\
  --input-features, -f\t\tFeatures file (default.f)\n\
  --input-dictionary, -d\tDictionary file (default.dict)\n\
  --input-alphas, -a\t\tAlphas file (default.alpha)\n\
  --prune, -p\t\t\tPrune: limit number of states (keep all)\n\
  --nbest, -b\t\t\tNBest: output n top-scoring solutions (1)\n\
  --nbest-score, -s\t\tNBest: show score for proposed solution (no)\n\
  --order, -o\t\t\tModel order (n-gram) (3)\n\
  --verbose, -v\t\t\tBe verbose\n\
  --help, -h\t\t\tThis help\n\
  --version, -V\t\t\tVersion information\n");
		break;
	case CMD_NULL:
	default:
		fprintf(stderr, "\
usage: perc <make_features|train|test> [options] data_file ...\n\
Use perc [command] --help to get info about options.\n");
	}
	exit(0);
}

void print_version() {
	fprintf(stderr, "perc (%dbit alpha), %s(%s)\nbugreports: %s\n", ALPHA_SIZE, PACKAGE_STRING, svnVersion, PACKAGE_BUGREPORT);
	exit(0);
}

int process_options(int argc, char *argv[]) {
	while (1) {
		int option_index = 0;
		static struct option long_options[] = {
			{ "feature-templates", 1, NULL, 't' },
			{ "input-dictionary", 1, NULL, 'd'},
			{ "output-dictionary", 1, NULL, 'D'},
			{ "input-features", 1, NULL, 'f' },
			{ "output-features", 1, NULL, 'F' },
			{ "input-alphas", 1, NULL, 'a' },
			{ "output-alphas", 1, NULL, 'A' },
			{ "min-freq", 1, NULL, 'n'},
			{ "iterations", 1, NULL, 'i'},
			{ "prune", 1, NULL, 'p'},
			{ "order", 1, NULL, 'o'},
			{ "nbest", 1, NULL, 'b'},
			{ "nbest-score", 0, NULL, 's'},
			{ "verbose", 0, NULL, 'v' },
			{ "help", 0, NULL, 'h' },
			{ "version", 0, NULL, 'V'},
			{ NULL, 0, NULL, 0 }
		};

		int c = getopt_long(argc, argv, "t:d:D:f:F:a:A:n:i:p:o:b:svhV", long_options, &option_index);
		if (c == -1)
			break;

		switch (c) {
		case 't':
			feature_templates = strdup(optarg);
			break;
		case 'd':
			input_dictionary = strdup(optarg);
			break;
		case 'D':
			output_dictionary = strdup(optarg);
			break;
		case 'f':
			input_features = strdup(optarg);
			break;
		case 'F':
			output_features = strdup(optarg);
			break;
		case 'a':
			input_alphas = strdup(optarg);
			break;
		case 'A':
			output_alphas = strdup(optarg);
			break;
		case 'n':
			min_freq = atoi(optarg);
			if (min_freq < 1)
				min_freq = 1;
			break;
		case 'i':
			iterations = atoi(optarg);
			break;
		case 'p':
			prune = atoi(optarg);
			break;
		case 'o':
			order = atoi(optarg);
			if (order < 2) {
				fprintf(stderr, "Warning: invalid model order, using 3-grams\n");
				order = 3;
			}
			break;
		case 'b':
			nbest = atoi(optarg);
			break;
		case 's':
			nbest_score = true;
			break;
		case 'v':
			verbose++;
			break;
		case 'h':
			help = true;
			break;
		case 'V':
			print_version();
			break;
		default:
			break;
		}
	}

	return optind;
}

int main(int argc, char *argv[]) {
	int cmdIndex = process_options(argc, argv);

	if (cmdIndex >= argc)
		print_help();
	if (!strcmp(argv[cmdIndex], "make_features")) {
		command = CMD_MAKE_FEATURES;
		if (!feature_templates)
			feature_templates = strdup("default.ft");
		if (!output_features)
			output_features = strdup("default.f");
		if (!output_dictionary)
			output_dictionary = strdup("default.dict");
	} else if (!strcmp(argv[cmdIndex], "train")) {
		command = CMD_TRAIN;
		if (!input_features)
			input_features = strdup("default.f");
		if (!input_dictionary)
			input_dictionary = strdup("default.dict");
		if (!output_alphas)
			output_alphas = strdup("default.alpha");
	} else if (!strcmp(argv[cmdIndex], "test")) {
		command = CMD_TEST;
		if (!input_features)
			input_features = strdup("default.f");
		if (!input_dictionary)
			input_dictionary = strdup("default.dict");
		if (!input_alphas)
			input_alphas = strdup("default.alpha");
	}
	if (help || command == CMD_NULL)
		print_help();
	
	int dataIndex = cmdIndex+1;
	if (dataIndex >= argc)
		print_help();

	char str[MAX_LINE_SIZE];
	FILE *f;

	f = fopen(argv[dataIndex], "r");
	if (!f)
		die("Cannot open file: %s", argv[dataIndex]);
	// read data file header
	errno = 0;
	if (!fgets(str, sizeof(str), f))
		die("Cannot read data header (%s)", argv[dataIndex]);

	PERC *perc = perc_init();
	switch (command) {
	case CMD_MAKE_FEATURES:
		if (!perc_make_features_init(perc, feature_templates, input_dictionary, str))
			exit(1);
		break;
	case CMD_TRAIN:
		if (!perc_train_init(perc, input_features, input_dictionary, input_alphas, str, prune, order))
			exit(1);
		break;
	case CMD_TEST:
		if (!perc_test_init(perc, input_features, input_dictionary, input_alphas, str, prune, nbest, order))
			exit(1);
		break;
	default:
		break;
	}

	if (verbose > 0)
		fprintf(stderr, "0 kw processed\n");

	int fileIndex = dataIndex;

	for (int iter = 0; iter < iterations; iter++) {
		int total = 0;
		int last = 0;
		bool firstSentence = true;

		while (fileIndex < argc) {
			if (!f) {
				f = fopen(argv[fileIndex], "r");
				if (!f)
					die("Cannot open file: %s", argv[fileIndex]);
				// skip data file header: TODO: check header (that it is compatible) instead!!!
				errno = 0;
				if (!fgets(str, sizeof(str), f))
					die("Cannot read data header (%s)", argv[fileIndex]);
			}
	
			// read entire file
			int eof = 0;
			int eos = 0;
			int sentenceSize = 0;
			perc_begin_sentence(perc);
			do {
				errno = 0;
				if (!fgets(str, sizeof(str), f)) {
					if (errno > 0)
						die("Cannot read from file: ");
					eof = 1;
				} else {
					chomp(str);
				}
				if (eof || !isEmpty(str)) {
					if (eos || eof) {
						double c, ac;
						perc_end_sentence(perc);
						switch (command) {
						case CMD_MAKE_FEATURES:
							if (verbose > 0) {
								if (last/10000 < total/10000) {
									fprintf(stderr, "%d kw processed\n", total/1000);
								}
								last = total;
							}
							break;
						case CMD_TRAIN:
							// print update
							if (verbose > 0) {
								if (last/10000 < total/10000) {
									perc_get_accuracy(perc, &c, &ac);
									fprintf(stderr, "%d kw processed, accuracy: %.3f", total/1000, 100*c);
									if (ac >= 0)
										fprintf(stderr, " (%.3f)", 100*ac);
									fputc('\n', stderr);
								}
								last = total;
							}
							break;
						case CMD_TEST:
							if (firstSentence) {
								time_t now;
								time(&now);
								printf("# Output%s created: %s", nbest_score ? " (score)" : "", ctime(&now));
								firstSentence = false;
							} else {
								putchar('\n');
							}

							// print proposed tags
							if (nbest_score) {
								for (int i = 0; i < nbest; i++) {
									if (i > 0)
										putchar('\t');
									printf("%.3lf", perc_get_nbest_score(perc, i));
								}
								putchar('\n');
							}
							for (int i = 0; i < sentenceSize; i++) {
								for (int j = 0; j < nbest; j++) {
									if (j > 0)
										putchar('\t');
									fputs(perc_get_proposed_tag(perc, i, j), stdout);
									const char *s = perc_get_additional_proposed_tag(perc, i, j);
									if (s) {
										putchar(' ');
										fputs(s, stdout);
									}
								}
								putchar('\n');
							}

							// print update
							if (verbose > 0) {
								if (last/10000 < total/10000) {
									perc_get_accuracy(perc, &c, &ac);
									fprintf(stderr, "%d kw processed, accuracy: %.3f", total/1000, 100*c);
									if (ac >= 0)
										fprintf(stderr, " (%.3f)", 100*ac);
									fputc('\n', stderr);
								}
								last = total;
							}
							break;
						default:
							break;
						}
						eos = 0;
						sentenceSize = 0;
						if (!eof)
							perc_begin_sentence(perc);
					}
					if (!eof) {
						if (perc_append_word(perc, str)) {
							total++;
							sentenceSize++;
						}
					}
				} else {
					eos = 1;
				}
			} while (!eof);
	
			fclose(f);
			f = NULL;
			fileIndex++;
		}

		double c, ac;
		switch (command) {
		case CMD_MAKE_FEATURES:
			if (verbose > 0)
				fprintf(stderr, "Words processed: %d\n", total);
			break;
		case CMD_TRAIN:
			if (verbose > 0) {
				perc_get_accuracy(perc, &c, &ac);
				perc_reset_accuracy(perc);
				fprintf(stderr, "Iteration: %d, words processed: %d, accuracy: %.3f", iter+1, total, 100*c);
				if (ac >= 0)
					fprintf(stderr, " (%.3f)", 100*ac);
				fputc('\n', stderr);
			}
			break;
		case CMD_TEST:
			perc_get_accuracy(perc, &c, &ac);
			fprintf(stderr, "%d words processed, accuracy: %.3f", total, 100*c);
			if (ac >= 0)
				fprintf(stderr, " (%.3f)", 100*ac);
			fputc('\n', stderr);
			break;
		default:
			break;
		}

		fileIndex = dataIndex;
	}

	switch (command) {
	case CMD_MAKE_FEATURES:
		if (!perc_make_features_finish(perc, output_features, output_dictionary, min_freq))
			exit(1);
		break;
	case CMD_TRAIN:
		if (!perc_train_finish(perc, output_dictionary, output_alphas))
			exit(1);
		break;
	case CMD_TEST:
		if (!perc_test_finish(perc))
			exit(1);
		break;
	default:
		break;
	}

	perc_destroy(perc);

	free(feature_templates);
	free(input_features);
	free(output_features);
	free(input_dictionary);
	free(output_dictionary);
	free(input_alphas);
	free(output_alphas);

	return 0;
}
